r"""
:class:`Holding`\s describe :class:`.Opinion`\s` attitudes toward :class:`.Rule`\s.

:class:`Holding`\s are text passages within :class:`.Opinion`\s
in which :class:`.Court` posits, or rejects, the validity of a
:class:`.Rule` within the :class:`.Court`\'s :class:`.Jurisdiction`,
or the :class:`.Court` asserts that the validity of the :class:`.Rule`
should be considered undecided.
"""

from __future__ import annotations
from copy import deepcopy

from itertools import chain
import operator
from typing import Any, Callable, Dict, Iterable, Iterator, List
from typing import Optional, Sequence, Tuple, TypeVar, Union

from dataclasses import dataclass, field

from anchorpoint import TextQuoteSelector
from legislice.enactments import Enactment

from nettlesome.terms import (
    Comparable,
    ContextRegister,
    Explanation,
    FactorMatch,
    Term,
    TermSequence,
    contradicts,
    new_context_helper,
)
from nettlesome.factors import Factor
from nettlesome.formatting import indented, wrapped
from nettlesome.groups import FactorGroup

from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule


@dataclass()
class Holding(Comparable):
    """
    An :class:`.Opinion`\'s announcement that it posits or rejects a legal :class:`.Rule`.

    Note that if an opinion merely says the court is not deciding whether
    a :class:`.Rule` is valid, there is no :class:`Holding`, and no
    :class:`.Rule` object should be created. Deciding not to decide
    a :class:`Rule`\'s validity is not the same thing as deciding
    that the :class:`.Rule` is undecided.

    :param rule:
        a statement of a legal doctrine about a :class:`.Procedure` for litigation.

    :param rule_valid:
        ``True`` means the :class:`.Rule` is asserted to be valid (or
        useable by a court in litigation). ``False`` means it's asserted
        to be invalid.

    :param decided:
        ``False`` means that it should be deemed undecided
        whether the :class:`.Rule` is valid, and thus can have the
        effect of overruling prior holdings finding the :class:`.Rule`
        to be either valid or invalid. Seemingly, ``decided=False``
        should render the ``rule_valid`` flag irrelevant.

    :param exclusive:
        if True, the stated rule is asserted to be the only way to establish
        the output that is the output of the rule.

    :param generic:
        if True, indicates that the specific attributes of this holding
        are irrelevant in the context of a different holding that is
        referencing this holding.
    """

    rule: Rule
    rule_valid: bool = True
    decided: bool = True
    exclusive: bool = False
    generic: bool = False
    absent: bool = False
    anchors: List[TextQuoteSelector] = field(default_factory=list)

    def __post_init__(self):
        if self.exclusive:
            if not self.rule_valid:
                raise NotImplementedError(
                    "The ability to state that it is not 'valid' to assert "
                    + "that a Rule is the 'exclusive' way to reach an output is "
                    + "not implemented, so 'rule_valid' cannot be False while "
                    + "'exclusive' is True. Try expressing this in another way "
                    + "without the 'exclusive' keyword."
                )
            if not self.decided:
                raise NotImplementedError(
                    "The ability to state that it is not 'decided' whether "
                    + "a Rule is the 'exclusive' way to reach an output is "
                    + "not implemented. Try expressing this in another way "
                    + "without the 'exclusive' keyword."
                )
            self.rule.procedure.valid_for_exclusive_tag()

    @classmethod
    def from_factors(
        self,
        outputs: FactorGroup,
        inputs: Optional[FactorGroup] = None,
        despite: Optional[FactorGroup] = None,
        enactments: Sequence[Enactment] = (),
        enactments_despite: Sequence[Enactment] = (),
        mandatory: bool = False,
        universal: bool = False,
        generic: bool = False,
        decided: bool = True,
        exclusive: bool = False,
        absent: bool = False,
    ):
        """Create new Holding without an existing Rule or Procedure."""
        procedure = Procedure(inputs=inputs, outputs=outputs, despite=despite)
        rule = Rule(
            procedure=procedure,
            enactments=enactments,
            enactments_despite=enactments_despite,
            mandatory=mandatory,
            universal=universal,
        )
        return Holding(rule=rule, generic=generic, decided=decided, exclusive=exclusive)

    @property
    def procedure(self):
        """Get Procedure from Rule."""
        return self.rule.procedure

    @property
    def despite(self):
        """Get Factors that specifically don't preclude application of the Holding."""
        return self.rule.procedure.despite

    @property
    def inputs(self):
        """Get inputs from Procedure."""
        return self.rule.procedure.inputs

    @property
    def outputs(self):
        """Get outputs from Procedure."""
        return self.rule.procedure.outputs

    @property
    def enactments(self):
        """Get Enactments required to apply the Holding."""
        return self.rule.enactments

    @property
    def enactments_despite(self):
        """Get Enactments that specifically don't preclude application of the Holding."""
        return self.rule.enactments_despite

    @property
    def recursive_terms(self) -> Dict[str, Term]:
        r"""
        Collect `self`'s :attr:`terms`, and their :attr:`terms`, recursively.

        :returns:
            a :class:`dict` (instead of a :class:`set`,
            to preserve order) of :class:`Term`\s.
        """
        return self.rule.recursive_terms

    @property
    def terms(self) -> TermSequence:
        r"""
        Call :class:`Procedure`\'s :meth:`~Procedure.terms` method.

        :returns:
            terms from ``self``'s :class:`Procedure`
        """
        return self.rule.procedure.terms

    def generic_terms_by_str(self) -> Dict[str, Comparable]:
        r"""
        Get :class:`.Factor`\s that can be replaced without changing ``self``\s meaning.

        :returns:
            generic :class:`.Factor`\s from ``self``'s :class:`Procedure`
        """
        return self.rule.generic_terms_by_str()

    @property
    def mandatory(self) -> bool:
        """Whether court "MUST" apply holding when it is applicable."""
        return self.rule.mandatory

    @property
    def universal(self) -> bool:
        """Whether holding is applicable in "ALL" cases where inputs are present."""
        return self.rule.universal

    def add_if_not_exclusive(self, other: Holding) -> Optional[Holding]:
        """Show how first Holding triggers second, assumed not to be "exclusive" way to reach result."""
        new_rule = self.rule + other.rule
        if new_rule is None:
            return None
        new_holding = deepcopy(self)
        new_holding.rule = new_rule
        return new_holding

    def add_enactment(self, enactment: Enactment) -> None:
        """Add enactment and sort self's Enactments."""
        self.rule.add_enactment(enactment)

    def add_enactment_despite(self, enactment: Enactment) -> None:
        """Add "despite" enactment and sort self's "despite" Enactments."""
        self.rule.add_enactment_despite(enactment)

    def add_holding(self, other: Holding) -> Optional[Holding]:
        """Show how first Holding triggers the second."""
        if not (self.decided and other.decided):
            raise NotImplementedError(
                "Adding is not implemented for Holdings that assert a Rule is not decided."
            )
        if not (self.rule_valid and other.rule_valid):
            raise NotImplementedError(
                "Adding is not implemented for Holdings that assert a Rule is not valid."
            )
        for self_holding in self.nonexclusive_holdings:
            for other_holding in other.nonexclusive_holdings:
                added = self_holding.add_if_not_exclusive(other_holding)
                if added is not None:
                    return added
        return None

    def __add__(self, other: Comparable) -> Optional[Union[Rule, Holding]]:
        """
        Create new Holding combining self and other into a single step, if possible.

        The Holdings can be combined only if the application of Holding ``self``
        necessarily provides all the required inputs for the application of ``other``.
        """
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            return self.add_holding(other)
        new_rule = self.rule + other
        if new_rule is None:
            return None
        result = deepcopy(self)
        result.rule = new_rule
        return result

    def _explanations_contradiction_of_holding(
        self, other: Holding, context: Explanation
    ) -> Iterator[Explanation]:
        for self_holding in self.nonexclusive_holdings:
            for other_holding in other.nonexclusive_holdings:
                for explanation in self_holding._contradicts_if_not_exclusive(
                    other_holding, context=context
                ):
                    explanation.reasons = [
                        HoldingMatch(
                            left=self_holding,
                            operation=contradicts,
                            right=other_holding,
                        )
                    ]
                    yield explanation

    def explanations_contradiction(
        self, other: Factor, context: ContextRegister = None
    ) -> Iterator[Explanation]:
        r"""
        Find context matches that would result in a contradiction with other.

        Works by testing whether ``self`` would imply ``other`` if
        ``other`` had an opposite value for ``rule_valid``.

        This method takes three main paths depending on
        whether the holdings ``self`` and ``other`` assert that
        rules are decided or undecided.

        A ``decided`` :class:`Rule` can never contradict
        a previous statement that any :class:`Rule` was undecided.

        If rule A implies rule B, then a holding that B is undecided
        contradicts a prior :class:`Rule` deciding that
        rule A is valid or invalid.

        :param other:
            The :class:`.Factor` to be compared to self. Unlike with
            :meth:`~Holding.contradicts`\, this method cannot be called
            with an :class:`.Opinion` for `other`.

        :returns:
            a generator yielding :class:`.ContextRegister`\s that cause a
            contradiction.
        """
        if not self.comparable_with(other):
            raise TypeError(f"Type Holding cannot be compared with type {type(other)}.")
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)
        if isinstance(other, Procedure):
            other = Rule(procedure=other)
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, self.__class__):
            yield from self._explanations_contradiction_of_holding(other, context)
        else:
            yield from other.explanations_contradiction(self)

    def _contradicts_if_not_exclusive(
        self, other: Holding, context: Explanation
    ) -> Iterator[Explanation]:
        if isinstance(other, Holding) and other.decided:
            if self.decided:
                yield from self._explanations_implies_if_not_exclusive(
                    other.negated(), context=context
                )
            else:
                yield from chain(
                    other._implies_if_decided(self, context=context),
                    other._implies_if_decided(self.negated(), context=context),
                )

    def _explanations_implies_if_not_exclusive(
        self, other: Factor, context: Explanation
    ) -> Iterator[Explanation]:

        if self.decided and other.decided:
            yield from self._implies_if_decided(other, context)

        # A holding being undecided doesn't seem to imply that
        # any other holding is undecided, except itself and the
        # negation of itself.

        # It doesn't seem that any holding being undecided implies
        # that any holding is decided, or vice versa.

        elif not self.decided and not other.decided:
            yield from chain(
                self.explanations_same_meaning(other, context),
                self.explanations_same_meaning(other.negated(), context),
            )

    def __ge__(self, other: Optional[Factor]) -> bool:
        return self.implies(other)

    def comparable_with(self, other: Any) -> bool:
        """Check if other can be compared to self for implication or contradiction."""
        if other and not isinstance(other, Comparable):
            return False
        return not isinstance(other, Factor)

    def implies(
        self, other: Optional[Comparable], context: ContextRegister = None
    ) -> bool:
        r"""
        Test for implication.

        See :meth:`.Procedure.implies_all_to_all`
        and :meth:`.Procedure.implies_all_to_some` for
        explanations of how ``inputs``, ``outputs``,
        and ``despite`` :class:`.Factor`\s affect implication.

        :param other:
            A :class:`Holding` to compare to self, or a :class:`.Rule` to
            convert into such a :class:`Holding` and then compare

        :returns:
            whether ``self`` implies ``other``
        """
        if other is None:
            return True
        return any(
            explanation is not None
            for explanation in self.explanations_implication(other, context)
        )

    def explanations_implication(
        self,
        other: Comparable,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Iterator[ContextRegister]:
        """Yield contexts that would cause self and other to have same meaning."""
        if not self.comparable_with(other):
            raise TypeError(f"Type Holding cannot be compared with type {type(other)}.")
        if not isinstance(context, Explanation):
            context = Explanation.from_context(context)
        if isinstance(other, (Rule, Procedure)):
            other = Holding(rule=other)
        if not isinstance(other, self.__class__):
            if context:
                context = context.reversed_context()
            if other.implied_by(self, context=context):
                yield context
        elif self.exclusive is other.exclusive is False:
            yield from self._explanations_implies_if_not_exclusive(
                other, context=context
            )
        else:
            yield from self.nonexclusive_holdings.explanations_implication(
                other.nonexclusive_holdings, context=context
            )

    def implied_by(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> bool:
        r"""
        Test if other implies self.

        This function is for handling implication checks for classes
        that don't know the structure of the :class:`Holding` class,
        such as :class:`.Fact` and :class:`.Rule`\.
        """
        if context:
            context = context.reversed()
        if isinstance(other, Rule):
            return Holding(rule=other).implies(self, context=context)
        return other.implies(self, context=context)

    def _implies_if_decided(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        r"""
        Test if ``self`` implies ``other`` if they're both decided.

        This is a partial version of the
        :meth:`Holding.__ge__` implication function.

        :returns:
            whether ``self`` implies ``other``, assuming that
            ``self.decided == other.decided == True`` and that
            ``self`` and ``other`` are both :class:`Holding`\s,
            although ``rule_valid`` can be ``False``.
        """

        if self.rule_valid and other.rule_valid:
            yield from self.rule.explanations_implication(other.rule, context)

        elif not self.rule_valid and not other.rule_valid:
            yield from other.rule.explanations_implication(self.rule, context)

        # Looking for implication where self.rule_valid != other.rule_valid
        # is equivalent to looking for contradiction.

        # If decided rule A contradicts B, then B also contradicts A

        else:
            yield from self.rule._explanations_contradiction(other.rule, context)

    def __len__(self):
        r"""
        Count generic :class:`.Factor`\s needed as context for this :class:`Holding`.

        :returns:
            the number of generic :class:`.Factor`\s needed for
            self's :class:`.Procedure`.
        """

        return len(self.rule.procedure)

    @property
    def inferred_from_exclusive(self) -> List[Holding]:
        r"""
        Yield :class:`Holding`\s that can be inferred from the "exclusive" flag.

        The generator will be empty if `self.exclusive` is False.
        """
        if self.exclusive:
            return [
                Holding(rule=modified_rule)
                for modified_rule in self.rule.get_contrapositives()
            ]
        return []

    def explanations_same_meaning(
        self, other: Factor, context: Optional[ContextRegister] = None
    ) -> Iterator[ContextRegister]:
        """Yield contexts that would cause self and other to have same meaning."""
        if (
            isinstance(other, self.__class__)
            and self.rule_valid == other.rule_valid
            and self.decided == other.decided
        ):
            yield from self.rule.explanations_same_meaning(other.rule, context)

    def negated(self):
        """Get new copy of ``self`` with an opposite value for ``rule_valid``."""
        result = deepcopy(self)
        result.rule_valid = not self.rule_valid
        result.exclusive = False
        return result

    @new_context_helper
    def new_context(self, changes: ContextRegister) -> Factor:
        """
        Create new :class:`Holding`, replacing keys of ``changes`` with values.

        :returns:
            a version of ``self`` with the new context.
        """
        return Holding(
            rule=self.rule.new_context(changes),
            rule_valid=self.rule_valid,
            decided=self.decided,
        )

    @property
    def nonexclusive_holdings(self) -> HoldingGroup:
        r"""Yield all :class:`.Holding`\s with `exclusive is False` implied by self."""
        if not self.exclusive:
            return HoldingGroup([self])
        nonexclusive_holding = deepcopy(self)
        nonexclusive_holding.exclusive = False
        holdings = [nonexclusive_holding] + self.inferred_from_exclusive
        return HoldingGroup(holdings)

    def set_inputs(self, factors: Sequence[Factor]) -> None:
        self.rule.set_inputs(factors)

    def set_despite(self, factors: Sequence[Factor]) -> None:
        self.rule.set_despite(factors)

    def set_outputs(self, factors: Sequence[Factor]) -> None:
        self.rule.set_outputs(factors)

    def set_enactments(self, enactments: Sequence[Enactment]) -> None:
        self.rule.set_enactments(enactments)

    def set_enactments_despite(self, enactments: Sequence[Enactment]) -> None:
        self.rule.set_enactments_despite(enactments)

    def _union_if_not_exclusive(
        self, other: Holding, context: ContextRegister
    ) -> Optional[Holding]:
        if self.decided is other.decided is False:
            if self.rule.implies(other.rule, context=context):
                return other
            if other.rule.implies(self.rule, context=context.reversed()):
                return self
            return None

        if not self.decided or not other.decided:
            return None
        if self.rule_valid != other.rule_valid:
            return None

        if self.rule_valid is False:
            # If a Rule with input A present is not valid
            # and a Rule with input A absent is also not valid
            # then a version of the Rule with input A
            # omitted is also not valid.
            raise NotImplementedError(
                "The union operation is not yet implemented for Holdings "
                "that assert a Rule is not valid."
            )

        new_rule = self.rule.union(other.rule, context=context)
        if not new_rule:
            return None
        result = deepcopy(self)
        result.rule = new_rule
        result.exclusive = False
        return result

    def _union_with_holding(
        self, other: Holding, context: ContextRegister
    ) -> Optional[Holding]:
        for self_holding in self.nonexclusive_holdings:
            for other_holding in other.nonexclusive_holdings:
                united = self_holding._union_if_not_exclusive(
                    other_holding, context=context
                )
                if united is not None:
                    return united
        return None

    def union(
        self, other: Union[Rule, Holding], context: Optional[ContextRegister] = None
    ) -> Optional[Holding]:
        """
        Infer a Holding from all inputs and outputs of self and other, in context.

        Creates a new Holding with all of the inputs and all of the outputs
        of both of the two original Holdings.

        However, you only get such a new Holding if it can be inferred by
        accepting the truth of the two original Holdings.

        If self contradicts() other, the operation returns None. Likewise, if
        the two original Holdings both have the value False for the parameter
        universal, the operation will return None if it’s possible that the
        “SOME” cases where one of the original Holdings applies don’t
        overlap with the “SOME” cases where the other applies.
        """
        context = context or ContextRegister()
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if not isinstance(other, Holding):
            raise TypeError
        return self._union_with_holding(other, context=context)

    def __or__(self, other: Union[Rule, Holding]) -> Optional[Holding]:
        """Infer a Holding from all inputs and outputs of self and other."""
        return self.union(other)

    def __str__(self):
        action = (
            "consider UNDECIDED"
            if not self.decided
            else ("ACCEPT" if self.rule_valid else "REJECT")
        )
        exclusive = (
            (
                f" that the EXCLUSIVE way to reach "
                f"{self.rule.outputs[0].short_string} is"
            )
            if self.exclusive
            else ""
        )
        rule_text = indented(str(self.rule))
        return wrapped(f"the Holding to {action}{exclusive}") + f"\n{rule_text}"


class HoldingMatch(FactorMatch):
    left: Holding
    operation: Callable
    right: Holding


class HoldingGroup(FactorGroup):
    term_class = Holding

    def __init__(self, holdings: Union[Sequence[Holding], Holding] = ()):
        if isinstance(holdings, Iterable):
            holdings = tuple(holdings)
        else:
            holdings = (holdings,)
        if any(not isinstance(holding, Holding) for holding in holdings):
            raise TypeError("All objects in HoldingGroup must be type Holding.")
        self.sequence = holdings

    def _explanations_implication_of_holding(
        self, other: Holding, context: Explanation
    ) -> Iterator[Explanation]:
        for self_holding in self.sequence:
            for result in self_holding.explanations_implication(other, context=context):
                yield result.with_match(
                    FactorMatch(left=self_holding, operation=operator.ge, right=other)
                )

    def explanations_implication(
        self,
        other: Comparable,
        context: Optional[Union[ContextRegister, Explanation]] = None,
    ) -> Iterator[Explanation]:
        if isinstance(other, Rule):
            other = Holding(rule=other)
        explanation = Explanation.from_context(
            context=context, current=self, incoming=other
        )
        if isinstance(other, Holding):
            yield from self._explanations_implication_of_holding(
                other=other, context=context
            )
        elif isinstance(other, self.__class__):
            yield from self.verbose_comparison(
                operation=operator.ge,
                still_need_matches=list(other),
                explanation=explanation,
            )

    def verbose_comparison(
        self,
        operation: Callable,
        still_need_matches: Sequence[Factor],
        explanation: Explanation,
    ) -> Iterator[Explanation]:
        r"""
        Find one way for two unordered sets of :class:`.Factor`\s to satisfy a comparison.

        All of the elements of `other` need to fit the comparison. The elements of
        `self` don't all need to be used.

        Only returns one answer, to prevent expensive fruitless searching.

        :param context:
            a mapping of :class:`.Factor`\s that have already been matched
            to each other in the recursive search for a complete group of
            matches. Usually starts empty when the method is first called.

        :param still_need_matches:
            :class:`.Factor`\s that need to satisfy the comparison
            :attr:`comparison` with some :class:`.Factor` of :attr:`available`
            for the relation to hold, and have not yet been matched.

        :param explanation:
            an :class:`.Explanation` showing which :class:`.Factor`\s listed in the
            FactorGroups were matched to each other, and also including a
            :class:`.ContextRegister`\.

        :yields:
            context registers showing how each :class:`.Factor` in
            ``need_matches`` can have the relation ``comparison``
            with some :class:`.Factor` in ``available_for_matching``,
            with matching context.
        """
        still_need_matches = list(still_need_matches)

        if not still_need_matches:
            yield explanation
        else:
            other_holding = still_need_matches.pop()
            for self_holding in self:
                if operation(self_holding, other_holding):
                    new_explanation = explanation.with_match(
                        FactorMatch(
                            left=self_holding,
                            operation=explanation.operation,
                            right=other_holding,
                        )
                    )
                    next_step = iter(
                        self.verbose_comparison(
                            still_need_matches=still_need_matches,
                            operation=operation,
                            explanation=new_explanation,
                        )
                    )
                    yield next(next_step)



from __future__ import annotations

from dataclasses import dataclass, field
import datetime
import operator
from typing import Iterable, Iterator, List
from typing import Optional, Sequence, Union

from anchorpoint.textselectors import TextQuoteSelector

from nettlesome.terms import Comparable, ContextRegister, Explanation
from nettlesome.factors import Factor
from authorityspoke.holdings import Holding, HoldingGroup
from authorityspoke.opinions import Opinion, TextLinkDict
from authorityspoke.rules import Rule


@dataclass
class CAPCitation:
    cite: str
    reporter: Optional[str] = None
    case_ids: List[int] = field(default_factory=list)


class Decision(Comparable):
    r"""
    A court decision to resolve a step in litigation.

    Uses the model of a judicial decision from
    the `Caselaw Access Project API <https://api.case.law/v1/cases/>`_.
    One of these records may contain multiple :class:`.Opinion`\s.

    Typically one record will contain all
    the :class:`~authorityspoke.opinions.Opinion`\s
    from one appeal, but not necessarily from the whole lawsuit. One
    lawsuit may contain multiple appeals or other petitions, and
    if more then one of those generates published Opinions,
    the CAP API will divide those Opinions into a separate
    record for each appeal.

    The outcome of a decision may be determined by one majority
    :class:`~authorityspoke.opinions.Opinion` or by the combined
    effect of multiple Opinions.
    The lead opinion is commonly, but not always, the only
    Opinion that creates binding legal authority.
    Usually every :class:`~authorityspoke.rules.Rule` posited by the lead Opinion is
    binding, but some may not be, often because parts of the
    Opinion fail to command a majority of the panel
    of judges.


    :param name:
        full name of the opinion, e.g. "ORACLE AMERICA, INC.,
        Plaintiff-Appellant, v. GOOGLE INC., Defendant-Cross-Appellant"
    :param name_abbreviation:
        shorter name of the opinion, e.g. "Oracle America, Inc. v. Google Inc."
    :param citations:
        citations to the opinion, usually in the format
        ``[Volume Number] [Reporter Name Abbreviation] [Page Number]``
    :param first_page:
        the page where the opinion begins in its official reporter
    :param last_page:
        the page where the opinion ends in its official reporter
    :param decision_date:
        date when the opinion was first published by the court
        (not the publication date of the reporter volume)
    :param court:
        name of the court that published the opinion
    :param _id:
        unique ID from CAP API
    """

    def __init__(
        self,
        date: datetime.date,
        name: Optional[str] = None,
        name_abbreviation: Optional[str] = None,
        citations: Optional[Sequence[CAPCitation]] = None,
        first_page: Optional[int] = None,
        last_page: Optional[int] = None,
        court: Optional[str] = None,
        opinions: Optional[Union[Opinion, Sequence[Opinion]]] = None,
        jurisdiction: Optional[str] = None,
        cites_to: Optional[Sequence[CAPCitation]] = None,
        id: Optional[int] = None,
    ) -> None:
        self.date = date
        self.name = name
        self.name_abbreviation = name_abbreviation
        self.citations = citations
        self.first_page = first_page
        self.last_page = last_page
        self.court = court
        if isinstance(opinions, Opinion):
            self.opinions = [opinions]
        elif not opinions:
            self.opinions = []
        else:
            self.opinions = list(opinions)
        self.jurisdiction = jurisdiction
        self.cites_to = cites_to
        self._id = id

    def __str__(self):
        citation = self.citations[0].cite if self.citations else ""
        name = self.name_abbreviation or self.name
        return f"{name}, {citation} ({self.date})"

    @property
    def holdings(self) -> HoldingGroup:
        if self.majority is None:
            return HoldingGroup()
        return HoldingGroup(self.majority.holdings)

    def contradicts(self, other):
        if isinstance(other, Decision):
            if self.majority and other.majority:
                return self.majority.contradicts(other.majority)
            return False
        return self.majority.contradicts(other)

    def explain_contradiction(
        self, other: Union[Opinion, Holding, Rule]
    ) -> Optional[Explanation]:
        explanations = self.explanations_contradiction(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_contradiction(
        self,
        other: Union[Decision, Opinion, Holding, Rule],
    ) -> Iterator[Explanation]:
        if isinstance(other, Decision):
            if self.majority and other.majority:
                yield from self.majority.explanations_contradiction(other.majority)
        elif isinstance(other, (Rule, Holding, Opinion)):
            if self.majority:
                yield from self.majority.explanations_contradiction(other)
        else:
            raise TypeError(
                f"'Contradicts' test not implemented for types "
                f"{self.__class__} and {other.__class__}."
            )

    def explain_implication(
        self,
        other: Union[Opinion, Holding, Rule],
    ) -> Optional[Explanation]:
        explanations = self.explanations_implication(other)
        try:
            explanation = next(explanations)
        except StopIteration:
            return None
        return explanation

    def explanations_implication(
        self, other: Union[Decision, Opinion, Holding, Rule]
    ) -> Iterator[Explanation]:
        if isinstance(other, Decision):
            if self.majority and other.majority:
                yield from self.majority.explanations_implication(other.majority)
        elif isinstance(other, (Rule, Holding, Opinion)):
            if self.majority:
                yield from self.majority.explanations_implication(other)
        else:
            raise TypeError(
                f"'Implication' test not implemented for types "
                f"{self.__class__} and {other.__class__}."
            )

    def posit(
        self,
        holdings: Union[Holding, Iterable[Union[Holding, Rule]]],
        holding_anchors: Optional[
            List[Union[TextQuoteSelector, List[TextQuoteSelector]]]
        ] = None,
        named_anchors: Optional[TextLinkDict] = None,
        context: Optional[Sequence[Factor]] = None,
    ) -> None:
        """
        Add one or more Holdings to the majority Opinion of this Decision.
        """
        if self.majority is None:
            raise AttributeError(
                "Cannot posit Holding because this Decision has no known majority Opinion."
                " Try having an Opinion posit the Holding directly."
            )
        self.majority.posit(
            holdings=holdings,
            holding_anchors=holding_anchors,
            named_anchors=named_anchors,
            context=context,
        )

    def __ge__(self, other) -> bool:
        return self.implies(other)

    def __gt__(self, other) -> bool:
        return self.implies(other) and not self == other

    def implied_by_holding(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> Iterator[Explanation]:
        if all(
            other.implies(self_holding, context=context)
            for self_holding in self.holdings
        ):
            yield Explanation(
                reasons=[(other, self_holding) for self_holding in self.holdings],
                operation=operator.ge,
            )

    def explanations_implied_by(
        self, other: Comparable, context: Optional[ContextRegister] = None
    ) -> Iterator[Explanation]:
        context = context or ContextRegister()
        if isinstance(other, Opinion):
            other = other.holdings
        if isinstance(other, HoldingGroup):
            yield from other.explanations_implication(
                self.holdings, context=context.reversed()
            )
        if isinstance(other, Rule):
            other = Holding(rule=other)
        if isinstance(other, Holding):
            yield from self.implied_by_holding(other, context)

    def implied_by(
        self, other: Optional[Comparable], context: Optional[ContextRegister] = None
    ) -> bool:
        if other is None:
            return False
        return any(self.explanations_implied_by(other, context))

    def implies_holding(
        self, other: Holding, context: Optional[ContextRegister] = None
    ) -> bool:
        return any(
            self_holding.implies(other, context=context)
            for self_holding in self.holdings
        )

    def implies_rule(
        self, other: Rule, context: Optional[ContextRegister] = None
    ) -> bool:
        return self.implies_holding(Holding(other), context=context)

    def implies(self, other, context: Optional[ContextRegister] = None) -> bool:
        if isinstance(other, (Decision, Opinion)):
            return self.holdings.implies(other.holdings)
        elif isinstance(other, Holding):
            return self.implies_holding(other, context=context)
        elif isinstance(other, Rule):
            return self.implies_rule(other, context=context)
        return False

    @property
    def majority(self) -> Optional[Opinion]:
        for opinion in self.opinions:
            if opinion.position == "majority":
                return opinion
        return None



from copy import deepcopy
from datetime import date
import os
from typing import Type
from legislice.enactments import Enactment

import pytest

from dotenv import load_dotenv
from legislice.download import Client

from authorityspoke.facts import Fact
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule
from authorityspoke.holdings import Holding

from nettlesome.terms import ContextRegister, TermSequence
from nettlesome.entities import Entity
from nettlesome.predicates import Predicate
from nettlesome.statements import Statement
from nettlesome.groups import FactorGroup


load_dotenv()

TOKEN = os.getenv("LEGISLICE_API_TOKEN")
legislice_client = Client(api_token=TOKEN)


class TestHolding:
    def test_complex_string(self, make_complex_rule):
        holding = Holding(make_complex_rule["accept_murder_fact_from_relevance"])
        string = " ".join(x.strip() for x in str(holding).splitlines())
        assert "is relevant to show the fact that <Alice>" in string.replace("/n", " ")

    def test_string_indentation(self, make_opinion_with_holding):
        """
        Test that the text of an Evidence string is indented even
        when it appears in a Holding.
        """
        lotus = make_opinion_with_holding["lotus_majority"]
        assert "    OF:\n" in str(lotus.holdings[2])

    def test_repr(self, make_holding):
        assert "rule=Rule(" in repr(make_holding["h1"])

    def test_line_break_in_fact_within_holding(self, make_opinion_with_holding):
        """
        Test that holding uses the Fact string method with line breaks.
        """
        lotus = make_opinion_with_holding["lotus_majority"]
        assert "registered a copyright\n" in str(lotus.holdings[2])

    def test_case_class_name_for_fact_within_holding(self, make_opinion_with_holding):
        lotus = make_opinion_with_holding["lotus_majority"]
        assert "the Fact that <Lotus" not in str(lotus.holdings[2])
        assert "the fact that <Lotus" in str(lotus.holdings[2])

    def test_holding_without_inputs_not_exclusive(self, make_factor):
        with pytest.raises(ValueError):
            Holding(Rule(Procedure(outputs=make_factor["f_no_crime"])), exclusive=True)

    def test_holding_with_absent_output_not_exclusive(self, make_exhibit):
        with pytest.raises(ValueError):
            Holding(
                Rule(Procedure(outputs=make_exhibit["reciprocal_testimony_absent"])),
                exclusive=True,
            )

    def test_holding_with_two_outputs_not_exclusive(self, make_factor):
        with pytest.raises(ValueError):
            Holding(
                Rule(
                    Procedure(
                        outputs=[make_factor["f_no_crime"], make_factor["f_shooting"]]
                    )
                ),
                exclusive=True,
            )

    def test_infer_from_exclusive(self, make_opinion_with_holding):
        """
        Test that the correct inference is made from a Holding being marked
        the "exclusive" way to reach the output "{} infringed the copyright in {}"
        """
        exclusive_holding = make_opinion_with_holding["lotus_majority"].holdings[0]
        inferred = exclusive_holding.inferred_from_exclusive[0]
        lower = "absence of the fact that <Borland International> infringed the copyright in <the Lotus menu command hierarchy>".lower()
        assert inferred.outputs[0].short_string.lower() == lower
        assert inferred.outputs[0].absent

    def test_infer_from_not_exclusive(self, make_holding):
        """
        For a Holding with `exclusive=False`, inferred_from_exclusive
        should return an empty list.
        """
        assert not make_holding["h1"].inferred_from_exclusive

    def test_type_of_terms(self, make_holding):
        assert isinstance(make_holding["h1"].terms, TermSequence)

    def test_get_evidence_by_name_from_holding(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        holding = watt.holdings[4]
        assert holding.outputs[0].exhibit.name == "proof of Wattenburg's guilt"
        factor = holding.get_factor_by_name("proof of Wattenburg's guilt")
        assert factor.name == "proof of Wattenburg's guilt"

    def test_enactment_text_in_holding_str(self, make_opinion_with_holding):
        lotus = make_opinion_with_holding["lotus_majority"]
        holding = lotus.holdings[2]
        assert holding.enactments[0].content.startswith("In any judicial")
        assert "In any judicial" in holding.enactments[0].selected_text()

    def test_holding_with_standard_of_proof(self, make_factor):
        holding = Holding.from_factors(
            inputs=make_factor["f_shooting_craig_brd"],
            outputs=make_factor["f_crime_craig_poe"],
        )
        assert (
            "by the STANDARD beyond reasonable doubt".lower()
            in holding.wrapped_string.lower()
        )

    def test_holding_applies_despite_enactments(
        self, make_holding, e_copyright_requires_originality
    ):
        holding = make_holding["h2_despite_due_process"]
        assert "due process" in str(holding)
        holding.set_enactments_despite(e_copyright_requires_originality)
        assert "due process" not in str(holding)
        assert "Copyright protection subsists" in str(holding)
        assert len(holding.enactments_despite) == 1

    def test_regulation_ordered_after_statute(
        self, make_holding, e_copyright_requires_originality
    ):
        holding = make_holding["h2"]
        regulation = Enactment(
            node="/us/cfr/t37/s202.1",
            heading="",
            start_date=date(1992, 2, 21),
            content="The following are examples of works not subject to copyright",
        )
        holding.add_enactment(regulation)
        holding.add_enactment(e_copyright_requires_originality)
        assert holding.enactments[-1].node == "/us/cfr/t37/s202.1"

    def test_set_blank_despite(self, make_holding):
        holding = make_holding["h2"]
        assert len(holding.despite) == 1
        holding.set_despite([])
        assert len(holding.despite) == 0

    def test_change_outputs(self, make_holding, watt_factor):
        holding = make_holding["h2"]
        assert len(holding.outputs) == 1
        holding.set_outputs([watt_factor["f11"]])
        assert len(holding.outputs) == 1

    def test_cannot_be_exclusive_and_invalid(self, make_rule):
        with pytest.raises(NotImplementedError):
            Holding(rule=make_rule["h1"], exclusive=True, rule_valid=False)

    def test_cannot_be_exclusive_and_undecided(self, make_rule):
        with pytest.raises(NotImplementedError):
            Holding(rule=make_rule["h1"], exclusive=True, decided=False)


class TestSameMeaning:
    def test_identical_holdings_equal(self, make_holding):
        assert make_holding["h1"].means(make_holding["h1_again"])

    def test_holding_does_not_mean_None(self, make_holding):
        assert not make_holding["h1"].means(None)

    def test_negated_method(self, make_holding):
        assert make_holding["h1"].negated().means(make_holding["h1_opposite"])


class TestImplication:
    def test_undecided_holding_no_implication_more_inputs(self, make_holding):

        """h2 beind undecided doesn't imply that a version of
        h2 with more supporting factors is undecided"""

        assert (
            not make_holding["h2_undecided"]
            >= make_holding["h2_irrelevant_inputs_undecided"]
        )

    def test_undecided_holding_no_implication_fewer_inputs(self, make_holding):

        """h2_irrelevant_inputs being undecided does not imply that h2
        is undecided. If courts SOMEtimes MAY use the procedure in h2,
        it may or may not be decided whether any court has been allowed
        to apply h2_irrelevant_inputs, even though it has all of h2's
        supporting factors and no more undercutting factors.
        """

        assert not (
            make_holding["h2_irrelevant_inputs_undecided"]
            >= make_holding["h2_undecided"]
        )

    def test_no_undecided_holding_implication_by_MUST(self, make_holding):

        """If it's undecided whether courts MUST follow the procedure in h2,
        it still could be decided that they MAY do so"""

        assert not make_holding["h2_MUST_undecided"] >= make_holding["h2_undecided"]

    def test_no_undecided_holding_implication_of_MUST(self, make_holding):

        """If it's undecided whether courts MAY follow the procedure in h2,
        the rule that they MUST do so still could have been decided to be not valid."""

        assert not make_holding["h2_undecided"] >= make_holding["h2_MUST_undecided"]

    def test_no_undecided_holding_implication_with_ALL(self, make_holding):

        """If it's undecided whether courts ALWAYS MAY follow the procedure in h2,
        it still could be decided (in the negative) whether they ALWAYS MAY
        follow a version with fewer supporting inputs."""

        assert not (
            make_holding["h_near_means_curtilage_ALL_undecided"]
            >= make_holding["h2_undecided"]
        )

    def test_undecided_implies_negation_is_undecided(self, make_holding):
        assert make_holding["h2_invalid_undecided"] >= make_holding["h2_undecided"]
        assert make_holding["h2_undecided"] >= make_holding["h2_invalid_undecided"]

    def test_no_implication_undecided_to_decided(self, make_holding):
        assert not make_holding["h2_undecided"] >= make_holding["h2"]

    def test_no_implication_decided_to_undecided(self, make_holding):
        assert not make_holding["h2"] > make_holding["h2_invalid_undecided"]

    def test_no_implication_of_procedure(self, make_holding, make_procedure):
        assert not make_holding["h2_undecided"] >= make_procedure["c2"]

    def test_holding_implies_opinion_with_no_holdings(
        self, make_opinion_with_holding, make_opinion
    ):
        lotus = make_opinion["lotus_majority"]
        holding = make_opinion_with_holding["oracle_majority"].holdings[0]
        context = ContextRegister()
        context.insert_pair(
            Entity("the Java API"), Entity("the Lotus menu command hierarchy")
        )
        assert holding.implies(lotus, context=context)

    def test_holding_implies_none(self, make_holding):
        assert make_holding["h3"] >= None

    def test_holding_implies_rule(self, make_holding, make_rule):
        assert make_holding["h3"] >= make_rule["h3"]

    def test_implication_due_to_always(self, make_decision_with_holding):
        """
        A "MUST ALWAYS" Holding about copyrightability implies that
        things "MAY SOMETIMES" be copyrightable if they're computer
        programs. Makes sense in this context, but it would also
        have found implication of "Socrates is human, thus Socrates
        may sometimes be copyrightable."

        This goes back to the issue that new Holdings can't be
        generated just because they'd be implied if they existed.
        """
        oracle = make_decision_with_holding["oracle"]
        assert oracle.holdings[18].implies(oracle.holdings[19])
        assert oracle.holdings[19].implied_by(oracle.holdings[18])

    def test_implication_by_exclusive_holding(
        self, make_decision_with_holding, e_copyright_protection
    ):
        lotus_holding = make_decision_with_holding["lotus"].holdings[0]
        nonexclusive = Holding.from_factors(
            inputs=FactorGroup(
                [
                    Fact(
                        "${work} was copyrightable", terms=Entity("the birthday song")
                    ),
                    Fact(
                        "$person copied constituent elements of $work that were original",
                        terms=[Entity("Eve"), Entity("the birthday song")],
                    ),
                ]
            ),
            outputs=Fact(
                "$person infringed the copyright in $work",
                terms=[Entity("Eve"), Entity("the birthday song")],
            ),
            enactments=e_copyright_protection,
        )
        assert lotus_holding.implies(nonexclusive)

    def test_explanation_same_generic_factor(self, make_decision_with_holding):
        """
        Test that the ContextRegister makes sense when the same generic
        Factors occur in Holdings on both sides of the "implies" relationship.
        """
        oracle = make_decision_with_holding["oracle"]
        context = ContextRegister()
        language = Entity("the Java language")

        context.insert_pair(Entity("the Java API"), language)

        new_context = oracle.holdings[18].new_context(context)
        explanation = new_context.explain_implication(oracle.holdings[19])
        assert explanation.context.get_factor(language).compare_keys(language)

    def test_not_implied_by_statement(self, make_holding):
        assert not Statement(
            Predicate("$person was a person"), terms=Entity("Alice")
        ).implies(make_holding["h1"])

    def test_cannot_check_if_holding_implies_factor(self, make_holding, make_factor):
        with pytest.raises(TypeError):
            make_holding["h1"].implies(make_factor["f_no_crime"])

    def test_cannot_check_if_holding_implies_string(self, make_holding):
        with pytest.raises(TypeError):
            make_holding["h1"].implies("f_no_crime")


class TestContradiction:
    def test_holding_contradicts_invalid_version_of_self(self, make_holding):
        assert make_holding["h2"].negated().means(make_holding["h2_invalid"])
        assert make_holding["h2"].contradicts(make_holding["h2_invalid"])
        assert make_holding["h2"] >= make_holding["h2_invalid"].negated()

    def test_contradicts_if_valid(self, make_holding):

        assert make_holding["h2_ALL"].contradicts(
            make_holding["h2_SOME_MUST_output_false"]
        )
        assert make_holding["h2_ALL"].rule.contradicts(
            make_holding["h2_SOME_MUST_output_false"].rule
        )

    def test_contradicts_if_valid_invalid_holding(self, make_holding):

        """
        In the current design, Holding.contradicts calls implies;
        implies calls Rule.contradicts.
        """

        assert make_holding["h2_invalid"].contradicts(
            make_holding["h2_irrelevant_inputs"]
        )
        assert not make_holding["h2_invalid"].rule.contradicts(
            make_holding["h2_irrelevant_inputs"].rule
        )

    def test_negation_of_h2_contradicts_holding_that_implies_h2(self, make_holding):
        assert make_holding["h2_invalid"].contradicts(
            make_holding["h2_irrelevant_inputs"]
        )
        assert make_holding["h2_irrelevant_inputs"].contradicts(
            make_holding["h2_invalid"]
        )

    def test_invalid_holding_contradicts_h2(self, make_holding):
        """
        You NEVER MAY follow X
        will contradict
        You SOMEtimes MAY follow Y
        if X implies Y
        """
        assert make_holding["h2_invalid"].contradicts(
            make_holding["h2_irrelevant_inputs"]
        )
        assert make_holding["h2_irrelevant_inputs"].contradicts(
            make_holding["h2_invalid"]
        )

    def test_holding_that_implies_h2_contradicts_negation_of_h2(self, make_holding):
        """
        Tests whether "contradicts" works reciprocally in this case.
        It should be reciprocal in every case so far, but maybe not for 'decided.'
        """

        assert make_holding["h2_ALL"].contradicts(
            make_holding["h2_SOME_MUST_output_false"]
        )
        assert make_holding["h2_SOME_MUST_output_false"].contradicts(
            make_holding["h2_ALL"]
        )

    def test_invalidity_of_implying_holding_contradicts_implied(self, make_holding):

        # You NEVER MUST follow X
        # will contradict
        # You SOMEtimes MUST follow Y
        # if X implies Y or Y implies X

        assert make_holding["h2_MUST_invalid"].contradicts(
            make_holding["h2_irrelevant_inputs_MUST"]
        )
        assert make_holding["h2_irrelevant_inputs_MUST"].contradicts(
            make_holding["h2_MUST_invalid"]
        )

    def test_contradiction_with_ALL_MUST_and_invalid_SOME_MUST(self, make_holding):

        # You ALWAYS MUST follow X
        # will contradict
        # You NEVER MUST follow Y
        # if X implies Y or Y implies X

        assert make_holding["h2_ALL_MUST"].contradicts(
            make_holding["h2_irrelevant_inputs_MUST_invalid"]
        )
        assert make_holding["h2_irrelevant_inputs_MUST_invalid"].contradicts(
            make_holding["h2_ALL_MUST"]
        )

    def test_contradiction_with_ALL_MUST_and_invalid_ALL_MAY(self, make_holding):

        # You ALWAYS MUST follow X
        # will contradict
        # You MAY NOT ALWAYS follow Y
        # if Y implies X

        assert make_holding["h2_ALL_MUST"].contradicts(make_holding["h2_ALL_invalid"])
        assert make_holding["h2_ALL_invalid"].contradicts(make_holding["h2_ALL_MUST"])

    def test_contradiction_with_distance(self, make_opinion_with_holding, make_holding):
        watt = make_opinion_with_holding["watt_majority"]
        must_not_rule = make_holding["h2_output_false_ALL_MUST"]
        assert list(watt.holdings)[1].contradicts(must_not_rule)

    def test_holding_contradicts_opinion(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        context = ContextRegister()
        context.insert_pair(
            Entity("the Lotus menu command hierarchy"), Entity("the Java API")
        )
        assert lotus.holdings[6].contradicts(
            oracle,
            context=context,
        )

    def test_explain_holding_contradicting_opinion(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = lotus.holdings[6].explain_contradiction(oracle)
        assert (
            "<the java api> is like <the lotus menu command hierarchy>"
            in str(explanation).lower()
        )

    def test_no_holding_contradiction_explanations(self, make_opinion_with_holding):
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = lotus.holdings[1].explain_contradiction(lotus.holdings[2])
        assert explanation is None

    def test_no_comparison_holding_to_fact(self, make_holding, watt_factor):
        with pytest.raises(TypeError):
            make_holding["h1"].contradicts(watt_factor["f1"])

    def test_holding_does_not_contradict_fact(self, make_holding, watt_factor):
        assert not watt_factor["f1"].contradicts(make_holding["h1"])

    def test_error_no_contradiction_test(self, make_holding):
        with pytest.raises(TypeError):
            _ = make_holding["h1"].contradicts(ContextRegister({}))

    def test_contradiction_with_ALL_MUST_and_false_output_ALL_MAY(self, make_holding):

        # You ALWAYS MUST follow X
        # will contradict
        # You MAY NOT ALWAYS follow Y
        # if Y implies X

        assert make_holding["h2_ALL_MUST"].contradicts(
            make_holding["h2_output_false_ALL"]
        )
        assert make_holding["h2_output_false_ALL"].contradicts(
            make_holding["h2_ALL_MUST"]
        )

    def test_no_contradiction_when_added_enactment_makes_rule_valid(self, make_holding):
        assert not make_holding["h2_ALL_due_process"].contradicts(
            make_holding["h2_ALL_invalid"]
        )
        assert not make_holding["h2_ALL_invalid"].contradicts(
            make_holding["h2_ALL_due_process"]
        )

    def test_contradiction_with_fewer_enactments(self, make_holding):
        """This and the previous enactment contradiction test passed on their own.
        Does there need to be something about enactments in the contradicts method?"""
        assert make_holding["h2_ALL_due_process_invalid"].contradicts(
            make_holding["h2_ALL"]
        )

    # Contradiction of undecided Holdings

    def test_no_contradiction_same_undecided_holding(self, make_holding):
        assert not make_holding["h3_ALL_undecided"].contradicts(
            make_holding["h3_ALL_undecided"]
        )

    def test_undecided_contradicts_holding(self, make_holding):
        """When a lower court issues a holding deciding a legal issue
        and a higher court posits that the issue should be considered
        undecided, the lower court's prior holding is "contradicted"
        in the sense of being rendered ineffective."""

        assert make_holding["h2_undecided"].contradicts(make_holding["h2"])

    def test_undecided_contradicts_holding_reverse(self, make_holding):
        """
        Remember that the "contradicts" relation is not symmetric between
        decided and undecided Rules.
        """
        assert not make_holding["h2"].contradicts(make_holding["h2_undecided"])

    def test_undecided_contradicts_decided_invalid_holding(self, make_holding):
        assert make_holding["h2_undecided"].contradicts(make_holding["h2_invalid"])

    def test_no_contradiction_of_undecided_holding(self, make_holding):
        """A court's act of deciding a legal issue doesn't "contradict" another
        court's prior act of positing that the issue was undecided."""

        assert not make_holding["h2"].contradicts(make_holding["h2_undecided"])
        assert not make_holding["h2_invalid"].contradicts(make_holding["h2_undecided"])

    def test_undecided_holding_implied_contradiction(self, make_holding):
        assert make_holding["h2_irrelevant_inputs_undecided"].contradicts(
            make_holding["h2_ALL"]
        )
        assert not make_holding["h2_ALL"].contradicts(
            make_holding["h2_irrelevant_inputs_undecided"]
        )

    def test_undecided_holding_no_implied_contradiction_with_SOME(self, make_holding):
        """Because a SOME holding doesn't imply a version of the same holding
        with added supporting inputs,

        make_holding["h2_irrelevant_inputs_undecided"]
        does not contradict
        make_holding["h2"]

        which seems questionable."""

        assert not make_holding["h2_irrelevant_inputs_undecided"].contradicts(
            make_holding["h2"]
        )
        assert not make_holding["h2"].contradicts(
            make_holding["h2_irrelevant_inputs_undecided"]
        )

    def test_undecided_holding_no_implied_contradiction(self, make_holding):
        assert not make_holding["h2_irrelevant_inputs_undecided"].contradicts(
            make_holding["h2_ALL_invalid"]
        )
        assert not make_holding["h2_ALL_invalid"].contradicts(
            make_holding["h2_irrelevant_inputs_undecided"]
        )

    def test_contradiction_with_evidence(self, make_holding):
        assert make_holding["h3_ALL_undecided"].contradicts(
            make_holding["h3_fewer_inputs_ALL"]
        )

    def test_no_contradiction_holding_with_evidence(self, make_holding):
        assert not make_holding["h3_fewer_inputs_ALL_undecided"].contradicts(
            make_holding["h3_ALL"]
        )

    # Contradiction of other types

    def test_contradiction_with_procedure(self, make_holding, make_procedure):
        """
        This test previously required a TypeError, but on second
        thought the command expresses a pretty clear intention to
        convert the Procedure to a Holding.
        """
        assert make_holding["h2_undecided"].contradicts(make_procedure["c2"])


class TestAddition:
    def test_adding_same_ALL_holdings_results_in_same(self, make_opinion_with_holding):
        brad = make_opinion_with_holding["brad_majority"]
        new = brad.holdings[0] + brad.holdings[0]
        assert new.means(brad.holdings[0])

    def test_adding_same_SOME_holdings_results_in_None(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        assert watt.holdings[0] + watt.holdings[0] is None

    def test_add_rule_to_holding(self, make_opinion_with_holding):
        lotus = make_opinion_with_holding["lotus_majority"]
        oracle = make_opinion_with_holding["oracle_majority"]
        rule_from_lotus = lotus.holdings[0].inferred_from_exclusive[0].rule
        new_holding = oracle.holdings[0] + rule_from_lotus
        output_strings = (
            "the fact it was false that <the Java API> was copyrightable",
            "absence of the fact that <Borland International> "
            + "infringed the copyright in <the Java API>",
        )
        for output in new_holding.outputs:
            assert output.short_string in output_strings

    def test_add_exclusive_holding(self, make_opinion_with_holding):
        """
        The Rule will be interpreted as a Holding and will be added to
        one of the nonexclusive Holdings that can be inferred from the
        exclusive Holding.
        """
        feist = make_opinion_with_holding["feist_majority"]
        new_holding = feist.holdings[10] + feist.holdings[3]
        output_strings = (
            "the fact it was false that <Rural's telephone listings> were an original work",
            "absence of the fact that <Rural's telephone listings> were copyrightable",
        )
        for output in new_holding.outputs:
            assert output.short_string in output_strings

    def test_universal_holding_that_cannot_be_added(self, make_rule):
        left = make_rule["h2_ALL"]
        right = make_rule["h3_ALL"]
        assert left + right is None

    def test_add_contradictory_factor(self, make_holding):
        holding = make_holding["h1"]
        factor = holding.inputs[0].negated()
        assert holding + factor is None

    def test_cannot_add_invalid(self, make_holding):
        with pytest.raises(NotImplementedError):
            make_holding["h2"] + make_holding["h2_invalid"]

    def test_cannot_add_undecided(self, make_holding):
        with pytest.raises(NotImplementedError):
            make_holding["h2_undecided"] + make_holding["h2"]


class TestUnion:
    def test_union_neither_universal(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        holdings = list(feist.holdings)
        assert (holdings[9] | holdings[7]) is None

    def test_union_and_addition_different(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        result_of_adding = feist.holdings[10] + feist.holdings[3]
        result_of_union = feist.holdings[10] | feist.holdings[3]
        assert not result_of_adding.means(result_of_union)

    def test_union_with_exclusive_flag(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        result_of_union = feist.holdings[10] | feist.holdings[3]
        assert isinstance(result_of_union, Holding)
        assert result_of_union.universal is False

    def test_union_of_mandatory_and_mandatory_is_mandatory(
        self, make_opinion_with_holding
    ):
        feist = make_opinion_with_holding["feist_majority"]
        result_of_union = feist.holdings[0] | feist.holdings[2]
        assert result_of_union.mandatory
        assert result_of_union.universal

    def test_no_union_with_opinion(self, make_holding, make_opinion_with_holding):
        holding = make_holding["h1"]
        feist = make_opinion_with_holding["feist_majority"]
        with pytest.raises(TypeError):
            _ = holding | feist

    def test_union_with_rule(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        rule = feist.holdings[3].inferred_from_exclusive[0].rule
        new_holding = feist.holdings[10] | rule
        assert isinstance(new_holding, Holding)

    def test_no_union_with_an_undecided_holding(self, make_holding):
        right = make_holding["h1"]
        left = deepcopy(right)
        left.decided = False
        assert left | right is None

    def test_union_with_two_undecided_holdings(self, make_holding):
        narrow_undecided = make_holding["h2_ALL_MUST"]
        narrow_undecided.decided = False
        broad_undecided = make_holding["h2"]
        broad_undecided.decided = False
        new = narrow_undecided | broad_undecided
        assert new == broad_undecided
        assert broad_undecided | narrow_undecided == broad_undecided

    def test_union_unrelated_undecided(self, make_holding):
        left = make_holding["h1"]
        left.decided = False
        right = make_holding["h2"]
        right.decided = False
        assert left | right is None



from copy import deepcopy
import logging
import pytest

from nettlesome.terms import ContextRegister, Explanation, TermSequence
from nettlesome.entities import Entity
from nettlesome.groups import FactorGroup
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison, Q_

from authorityspoke.facts import Fact
from authorityspoke.procedures import Procedure


class TestProcedures:
    def test_exception_for_wrong_type_for_procedure(self, make_predicate):
        with pytest.raises(TypeError):
            Procedure(inputs=make_predicate["p1"], outputs=make_predicate["p2"])

    def test_exception_for_wrong_type_in_tuple_for_procedure(self, make_predicate):
        with pytest.raises(TypeError):
            Procedure(inputs=(make_predicate["p1"]), outputs=(make_predicate["p2"]))

    def test_get_terms(self, make_procedure):
        # motel, watt
        assert len(make_procedure["c1"].generic_terms()) == 2
        # trees, motel
        assert len(make_procedure["c2"].generic_terms()) == 2

    def test_procedure_length(self, make_procedure):
        """Consider deleting Procedure.__len__() and this test."""
        assert len(make_procedure["c2"]) == 2
        assert len(make_procedure["c1"]) == 2

    def test_procedure_string_with_entities(self, make_procedure):
        assert "<Craig> performed at <circus>" in str(
            make_procedure["c2_irrelevant_inputs"]
        )
        assert "<Dan> performed at <circus>" in str(
            make_procedure["c2_irrelevant_inputs"]
        )

    def test_cannot_add_nonfactor_as_input(self, make_factor):
        with pytest.raises(TypeError):
            Procedure(inputs="factor name", outputs=make_factor["f_shooting"])

    def test_cannot_add_term_as_input(self, make_factor):
        with pytest.raises(TypeError):
            Procedure(inputs=Entity("Al"), outputs=make_factor["f_shooting"])

    def test_generic_terms(self, make_entity, make_procedure, make_evidence):
        """
        Finds that for factor f["f7"], it would be consistent with the
        other group of factors for f["f7"]'s two slots to be assigned
        (0, 1) or (1, 0).
        """
        e = make_entity
        factors = make_procedure["c3"].generic_terms()
        for factor in (
            e["motel"],
            e["tree_search"],
            e["trees"],
            e["watt"],
            make_evidence["crime_absent"],
        ):
            assert factor in factors

    def test_type_of_terms(self, make_procedure):
        assert isinstance(make_procedure["c3"].terms, TermSequence)

    def test_repr(self, make_procedure):
        rep = repr(make_procedure["c3"])
        assert "to_effect=Fact(predicate=Predicate(template='$person committed" in rep

    def test_entities_of_inputs_for_identical_procedure(
        self, watt_factor, make_procedure, watt_mentioned
    ):
        f = watt_factor
        c1 = make_procedure["c1"]
        c1_again = make_procedure["c1_again"]
        assert f["f1"] in c1.inputs
        assert f["f1"] in c1_again.inputs
        assert f["f1"].terms == (watt_mentioned[0],)
        assert f["f2"] in c1.inputs
        assert f["f2"] in c1_again.inputs
        assert f["f2"].terms == (watt_mentioned[1], watt_mentioned[0])


class TestProcedureSameMeaning:
    def test_procedure_equality(self, make_procedure):
        assert make_procedure["c1"].means(make_procedure["c1_again"])

    def test_procedure_equality_entity_order(self, make_procedure):
        assert make_procedure["c1"].means(make_procedure["c1_entity_order"])

    def test_still_equal_after_swapping_reciprocal_entities(
        self, make_procedure, caplog
    ):
        caplog.set_level(logging.DEBUG)
        assert make_procedure["c2"].means(make_procedure["c2_reciprocal_swap"])

    def test_unequal_after_swapping_nonreciprocal_entities(self, make_procedure):
        assert not make_procedure["c2"].means(make_procedure["c2_nonreciprocal_swap"])

    def test_same_meaning_no_context(self, make_procedure):
        assert make_procedure["c_no_context"].means(make_procedure["c_no_context"])

    def test_explain_same_meaning(self, make_procedure):
        left = make_procedure["c1_factor_and_entity_order"]
        right = make_procedure["c1"]
        explanation = left.explain_same_meaning(right)
        assert "<Wattenburg> is like <Hideaway Lodge>" in str(explanation)
        assert "<Hideaway Lodge> is like <Wattenburg>" in str(explanation)


class TestProcedureImplication:
    def test_entities_of_implied_inputs_for_implied_procedure(
        self, watt_factor, make_procedure
    ):
        f = watt_factor
        c1_easy = make_procedure["c1_easy"]
        c1_order = make_procedure["c1_entity_order"]
        assert f["f2"] in c1_easy.inputs
        assert f["f1"] not in c1_easy.inputs

    def test_factor_implication_with_exact_quantity(self, watt_factor, make_procedure):
        """This test is mostly to demonstrate the relationships
        between the Factors in the Procedures that will be
        tested below."""
        f = watt_factor
        assert f["f7"] in make_procedure["c2"].inputs
        assert f["f7"] not in make_procedure["c2_exact_quantity"].inputs
        assert f["f8_exact"] in make_procedure["c2_exact_quantity"].inputs
        assert f["f8_exact"] > f["f7"]
        assert not f["f7"] >= f["f8_exact"]

    def test_procedure_implication_with_exact_quantity(self, make_procedure):
        """This is meant to show that the function finds the "distance is
        exactly 25" factor in c2_exact, and recognizes that factor can imply
        the "distance is more than 20" factor in c2 if they have the same entities.
        """

        c2 = make_procedure["c2"]
        c2_exact_quantity = make_procedure["c2_exact_quantity"]
        assert c2_exact_quantity.implies(c2)

    def test_procedure_general_quantity_does_not_imply_exact(
        self, watt_factor, make_procedure
    ):

        c2 = make_procedure["c2"]
        c2_exact_quantity = make_procedure["c2_exact_quantity"]
        assert not c2_exact_quantity <= c2

    def test_implied_procedure_with_reciprocal_entities(self, make_procedure):
        """
        Because both procedures have a form of "the distance between $place1 and $place2 was"
        factor and those factors are reciprocal, the entities of one of them in reversed
        order can be used as the entities of the other, and one will still imply the other.
        (But if there had been more than two entities, only the first two would have been
        reversed.)
        """

        c2 = make_procedure["c2"]
        c2_reciprocal_swap = make_procedure["c2_reciprocal_swap"]
        assert c2.means(c2_reciprocal_swap)
        assert c2 >= c2_reciprocal_swap

    def test_entities_of_implied_quantity_outputs_for_implied_procedure(
        self, make_procedure
    ):
        """
        If c2_narrow was "self" and c2_broad was "other", the output of
        c2_broad (with f["f8_int"]) would be implied by the output of
        c2_narrow (with f["f8_higher_int"]).
        """

        c2_broad = make_procedure["c2_broad_output"]
        c2_narrow = make_procedure["c2_narrow_output"]

        assert c2_narrow > c2_broad

    def test_procedure_implies_identical_procedure(self, make_procedure):
        assert make_procedure["c1"] >= make_procedure["c1_again"]
        assert make_procedure["c1"].means(make_procedure["c1_again"])

    def test_procedure_implies_same_procedure_fewer_inputs(
        self, make_procedure, caplog
    ):
        caplog.set_level(logging.DEBUG)
        assert make_procedure["c1_easy"] < make_procedure["c1"]
        assert make_procedure["c1_easy"] <= make_procedure["c1"]
        assert make_procedure["c1_easy"] != make_procedure["c1"]

    def test_procedure_implies_reordered_entities_fewer_inputs(self, make_procedure):

        assert make_procedure["c1_entity_order"] > make_procedure["c1_easy"]
        assert make_procedure["c1_easy"] < make_procedure["c1_entity_order"]
        assert make_procedure["c1_easy"] != make_procedure["c1_entity_order"]

    def test_procedure_exact_quantity_in_despite_implication(self, make_procedure):
        assert make_procedure["c2_exact_in_despite"] > make_procedure["c2"]

    def test_procedure_implication_despite_irrelevant_factors(self, make_procedure):
        assert make_procedure["c2"] < make_procedure["c2_irrelevant_inputs"]

    def test_implies_all_to_some(self, make_procedure):
        assert make_procedure["c2"].implies_all_to_some(
            make_procedure["c2_irrelevant_despite"]
        )
        assert not make_procedure["c2"] > make_procedure["c2_irrelevant_despite"]

    def test_all_to_some_implies_input_same_as_despite_of_other(self, make_procedure):
        """
        Every input of c2_exact_in_despite is equal to or implied by
        some input of c2, and an input of c2 implies the despite of c2_exact_in_despite.
        """
        p = make_procedure
        assert p["c2_exact_in_despite"].implies_all_to_some(p["c2"])

    def test_no_all_to_some_implies_input_contradicts_despite(self, make_procedure):
        """
        c2_higher_quantity has the right inputs, but it also has an
        input that contradicts the despite factor of c2_exact_in_despite.
        """
        p = make_procedure
        assert not p["c2_higher_quantity"].implies_all_to_some(p["c2_exact_in_despite"])

    def test_all_to_some_implication_added_despite_factors(self, make_procedure):

        assert not make_procedure["c2"].implies_all_to_some(
            make_procedure["c2_absent_despite"]
        )

    def test_implication_with_more_outputs_than_inputs(self, make_procedure):
        p = make_procedure
        assert p["c2_irrelevant_outputs"].implies_all_to_all(p["c2"])

    def test_fewer_inputs_implies_all_to_all(self, make_procedure):
        c = make_procedure
        assert c["c3_fewer_inputs"].implies_all_to_all(c["c3"])

    def test_all_to_all_implies_reciprocal(self, make_procedure, caplog):
        """
        These are the same Procedures below in
        test_implication_all_to_all_reciprocal
        """
        caplog.set_level(logging.DEBUG)
        assert make_procedure["c2_exact_in_despite_entity_order"].implies_all_to_all(
            make_procedure["c2"]
        )

    def test_no_implication_of_other_factor(self, make_procedure, watt_factor):
        assert not make_procedure["c1"].implies_all_to_all(watt_factor["f1"])
        assert not make_procedure["c1"].implies_all_to_some(watt_factor["f1"])


class TestProcedureContradiction:
    def test_no_contradict_between_procedures(self, make_procedure):
        """
        I don't think some-to-some contradiction is possible for Procedures
        """
        p = make_procedure
        with pytest.raises(NotImplementedError):
            assert p["c2_higher_quantity"].contradicts(p["c2_exact_in_despite"])

    def test_no_contradiction_of_other_type(self, make_procedure, watt_factor):
        assert not make_procedure["c1"].contradicts_some_to_all(watt_factor["f1"])

    def test_contradiction_some_to_all(self, watt_factor):
        within_curtilage = Procedure(
            inputs=(watt_factor["f9"],),
            outputs=watt_factor["f10"],
        )
        not_within_curtilage = Procedure(
            inputs=(watt_factor["f9"],),
            outputs=watt_factor["f10_false"],
        )
        assert not_within_curtilage.contradicts_some_to_all(within_curtilage)


class TestProcedureUnion:
    def test_simple_union(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        procedure_from_union = feist.holdings[0].procedure | feist.holdings[2].procedure
        procedure_from_adding = (
            feist.holdings[0].procedure + feist.holdings[2].procedure.inputs[0]
        )
        assert procedure_from_union.means(procedure_from_adding)


p_small_weight = Comparison(
    "the amount of gold $person possessed was", sign="<", expression=Q_("1 gram")
)
p_large_weight = Comparison(
    "the amount of gold $person possessed was",
    sign=">=",
    expression=Q_("100 kilograms"),
)
alice = Entity("Alice")
bob = Entity("Bob")
craig = Entity("Craig")
dan = Entity("Dan")
alice_rich = Fact(p_large_weight, terms=alice)
bob_poor = Fact(p_small_weight, terms=bob)
craig_rich = Fact(p_large_weight, terms=craig)
dan_poor = Fact(p_small_weight, terms=dan)


class TestFactorGroups:
    def test_consistent_factor_groups(self):
        """
        Verifies that the factor groups are considered "consistent"
        even though it would be possible to make an analogy that would
        make the statements contradict.
        """
        assert FactorGroup([alice_rich, bob_poor]).consistent_with(
            FactorGroup([dan_poor, craig_rich])
        )

    def test_consistent_factor_groups_with_context(self):
        alice_like_craig = ContextRegister()
        alice_like_craig.insert_pair(alice, craig)
        assert FactorGroup([alice_rich, bob_poor]).consistent_with(
            FactorGroup([dan_poor, craig_rich]),
            context=alice_like_craig,
        )

    def test_context_indicates_factors_match(self):
        alice_like_craig = ContextRegister()
        alice_like_craig.insert_pair(alice, craig)

        assert alice_rich.means(craig_rich, context=alice_like_craig)

    def test_inconsistent_factor_groups(self):
        """
        If Alice is considered analagous to Dan the two sets of
        statements would be inconsistent, but not if
        Alice is considered analagous to Craig.
        """
        alice_like_dan = ContextRegister()
        alice_like_dan.insert_pair(alice, dan)

        assert not FactorGroup([alice_rich, bob_poor]).consistent_with(
            FactorGroup([dan_poor, craig_rich]), context=alice_like_dan
        )

    def test_contradictory_factor_groups(self):
        """
        Verifies that the factor groups are considered "consistent"
        even though it would be possible to make an analogy that would
        make the statements contradict.

        If Alice is considered analagous to Dan the two sets of
        statements would be contradictory, but not if
        Alice is considered analagous to Craig.
        """
        assert FactorGroup([alice_rich, bob_poor]).contradicts(
            FactorGroup([craig_rich, dan_poor])
        )

    def test_not_contradictory_factor_groups(self):
        """
        Because the ContextRegister matches up the two contexts
        consistently, it's impossible to reach a contradiction.
        """
        alice_like_craig = ContextRegister()
        alice_like_craig.insert_pair(alice, craig)

        assert not FactorGroup((alice_rich, bob_poor)).contradicts(
            FactorGroup((dan_poor, craig_rich)),
            context=alice_like_craig,
        )

    def test_input_or_despite_factors(self):
        """Test part of the process of checking contradiction."""
        rural = Entity(name="Rural's telephone directory")
        compilation = Predicate(
            "${rural_s_telephone_directory} was a compilation of facts"
        )
        idea = Predicate(content="${rural_s_telephone_directory} was an idea")
        copyrightable = Fact(
            Predicate(content="${rural_s_telephone_directory} was copyrightable"),
            terms=rural,
        )
        left = Procedure(
            outputs=(copyrightable),
            inputs=(Fact(compilation, terms=rural)),
        )
        right = Procedure(
            outputs=(copyrightable),
            inputs=(Fact(idea, terms=rural)),
        )
        context = ContextRegister()
        explanation = Explanation.from_context(context)
        gen = left._has_input_or_despite_factors_implied_by_all_inputs_of(
            right, explanation
        )
        with pytest.raises(StopIteration):
            next(gen)


class TestEvolve:
    def test_evolve_context_to_absent(self, make_procedure):
        procedure = make_procedure["c1"]
        evolved = deepcopy(procedure)
        evolved.outputs[0].absent = True
        assert procedure.outputs[0].contradicts(evolved.outputs[0])



"""Functions to help locate data to import on disk."""

import pathlib

from typing import Optional


def make_filepath(
    filename: Optional[str] = None,
    directory: Optional[pathlib.Path] = None,
    filepath: Optional[pathlib.Path] = None,
    default_folder: str = "holdings",
):
    """
    Create :class:`.pathlib.Path` for a file containing XML or JSON to load as objects.

    :param filename:
        just the name of the file, without the directory.

    :param directory:
        just the directory, without the path

    :param filepath:
        a full path that just needs to be checked for validity and returned

    :param default_folder:
        a folder name to use in constructing a missing directory path

    :returns:
        the path to the desired file
    """
    if filepath:
        return pathlib.Path(filepath)
    if not directory:
        directory = get_directory_path(default_folder)
    to_append = filename or ""
    return directory / to_append


def get_directory_path(stem: str) -> pathlib.Path:
    """
    Find a data directory for importing files.

    Only in this module temporarily to prevent a circular import.

    :param stem:
        name of the folder where the desired example data files
        can be found, e.g. "holdings" or "opinions".

    :returns:
        path to the directory with the desired example data files.
    """
    directory = pathlib.Path.cwd()
    if directory.stem == stem:
        return directory
    if directory.stem != "example_data":
        directory = directory / "example_data"
    directory = directory / stem
    if not directory.exists():
        directory = pathlib.Path.cwd().parent / "example_data" / stem
    return directory



"""Nesting fields to prepare to load a dict with a Marshmallow schema."""

from typing import Callable, Dict, List, Sequence, Union


def nest_fields(data: Dict, nest: str, eggs: List[str]):
    """Make sure specified fields are nested under "nest" key."""
    for egg_field in eggs:
        if egg_field in data:
            if not data.get(nest):
                data[nest] = {}
            data[nest][egg_field] = data.pop(egg_field)
    return data


def walk_tree_and_modify(
    obj: Union[Dict, List], func: Callable, ignore: Sequence[str] = ()
) -> Union[Dict, List]:
    """
    Traverse tree of dicts and lists, and modify each node.

    :param obj: the object to traverse

    :param func:
        the function to call on each dict node, returning a dict

    :param ignore: the names of keys that should not be explored

    :returns: a version of the tree with every node modified by `func`
    """
    if isinstance(obj, List):
        return [walk_tree_and_modify(item, func, ignore) for item in obj]
    if isinstance(obj, Dict):

        obj_dict: Dict = func(obj)

        for key, value in obj_dict.items():
            if isinstance(value, (Dict, List)) and key not in ignore:
                obj_dict[key] = walk_tree_and_modify(value, func, ignore)

        return obj_dict

    return obj



import operator

import pytest

from nettlesome.terms import ContextRegister, means
from nettlesome.entities import Entity
from nettlesome.groups import FactorGroup
from nettlesome.statements import Statement


class TestContextRegisters:
    def test_possible_context_without_empty_spaces(self, watt_factor, make_entity):
        left = watt_factor["f1"]
        right = watt_factor["f1"]
        contexts = list(left.possible_contexts(right))
        assert len(contexts) == 1
        assert contexts[0].check_match(make_entity["motel"], make_entity["motel"])

    def test_all_possible_contexts_identical_factor(self, watt_factor, make_entity):
        left = watt_factor["f2"]
        right = watt_factor["f2"]
        contexts = list(left.possible_contexts(right))
        assert len(contexts) == 2
        assert any(
            context.check_match(make_entity["watt"], make_entity["motel"])
            for context in contexts
        )

    def test_context_not_equal_to_list(self):
        changes = ContextRegister.from_lists(
            [Entity("Alice")],
            [Entity("Dan")],
        )
        assert changes != [[Entity("Alice")], [Entity("Dan")]]

    def test_cannot_update_context_register_from_lists(self):
        left = Statement(
            "$shooter shot $victim", terms=[Entity("Alice"), Entity("Bob")]
        )
        right = Statement(
            "$shooter shot $victim", terms=[Entity("Craig"), Entity("Dan")]
        )
        update = left.update_context_register(
            right, context=[[Entity("Alice")], [Entity("Craig")]], comparison=means
        )
        with pytest.raises(TypeError):
            next(update)

    def test_limited_possible_contexts_identical_factor(self, watt_factor, make_entity):
        left = watt_factor["f2"]
        right = watt_factor["f2"]
        context = ContextRegister()
        context.insert_pair(make_entity["motel"], make_entity["motel"])
        contexts = list(left.possible_contexts(right, context=context))
        assert len(contexts) == 1
        assert contexts[0].check_match(make_entity["watt"], make_entity["watt"])

    def test_context_register_empty(self, make_complex_fact, watt_factor):
        """
        Yields no context_register because the Entity in f1 doesn't imply
        the Fact in f_relevant_murder.
        """
        with pytest.raises(StopIteration):
            next(
                watt_factor["f1"]._context_registers(
                    make_complex_fact["f_relevant_murder"], operator.ge
                )
            )

    def test_context_register_valid(self, make_entity, watt_factor):
        expected = ContextRegister()
        expected.insert_pair(make_entity["motel"], make_entity["watt"])
        generated = next(
            watt_factor["f1"]._context_registers(
                watt_factor["f1_entity_order"], operator.le
            )
        )
        assert generated == expected

    def test_import_to_context_register(self, make_entity, watt_factor):
        left = ContextRegister.from_lists(
            to_replace=[watt_factor["f7"], make_entity["motel"]],
            replacements=[watt_factor["f7_swap_entities"], make_entity["trees"]],
        )
        right = ContextRegister()
        right.insert_pair(make_entity["trees"], make_entity["motel"])
        assert len(left.merged_with(right)) == 3

    def test_import_to_mapping_no_change(self, make_entity):
        old_mapping = ContextRegister.from_lists(
            [make_entity["motel"]], [make_entity["trees"]]
        )
        new_mapping = ContextRegister.from_lists(
            [make_entity["motel"]], [make_entity["trees"]]
        )
        assert old_mapping.merged_with(new_mapping) == old_mapping

    def test_import_to_mapping_conflict(self, make_entity):
        old_mapping = ContextRegister.from_lists(
            [make_entity["motel"]], [make_entity["trees"]]
        )
        new_mapping = ContextRegister.from_lists(
            [make_entity["motel"]], [make_entity["motel"]]
        )
        merged = old_mapping.merged_with(new_mapping)
        assert merged is None

    def test_import_to_mapping_reciprocal(self, watt_factor):
        left = ContextRegister.from_lists([watt_factor["f7"]], [watt_factor["f7"]])
        right = ContextRegister.from_lists(
            [watt_factor["f7_swap_entities"]], [watt_factor["f7_swap_entities"]]
        )
        mapping = left.merged_with(right)
        assert mapping.get(str(watt_factor["f7"])).means(watt_factor["f7"])

    def test_registers_for_interchangeable_context(self, make_entity, watt_factor):
        """
        Test that _registers_for_interchangeable_context swaps the first two
        items in the ContextRegister
        """
        factor = watt_factor["f7"]
        first_pattern, second_pattern = list(factor.term_permutations())
        assert first_pattern[0] == second_pattern[1]
        assert first_pattern[1] == second_pattern[0]
        assert first_pattern[0] != first_pattern[1]


class TestLikelyContext:
    def test_likely_context_one_factor(self, make_entity, watt_factor):
        left = watt_factor["f2"]
        right = watt_factor["f2"]
        context = next(left.likely_contexts(right))
        assert context.check_match(make_entity["motel"], make_entity["motel"])

    def test_likely_context_implication_one_factor(self, make_entity, watt_factor):
        left = watt_factor["f8"]
        right = watt_factor["f8_meters"]
        context = next(left.likely_contexts(right))
        assert context.check_match(make_entity["motel"], make_entity["motel"])

    def test_likely_context_two_factors(self, make_entity, watt_factor):
        left = FactorGroup((watt_factor["f9_swap_entities"], watt_factor["f2"]))
        right = watt_factor["f2"]
        context = next(left.likely_contexts(right))
        assert context.check_match(make_entity["motel"], make_entity["motel"])

    def test_likely_context_two_by_two(self, make_entity, watt_factor):
        left = FactorGroup((watt_factor["f9"], watt_factor["f2"]))
        right = FactorGroup(
            (watt_factor["f9_swap_entities"], watt_factor["f9_more_different_entity"])
        )
        context = next(left.likely_contexts(right))
        assert context.check_match(make_entity["motel"], make_entity["trees"])

    def test_likely_context_different_terms(self, make_opinion_with_holding):
        lotus = make_opinion_with_holding["lotus_majority"]
        oracle = make_opinion_with_holding["oracle_majority"]
        left = [lotus.holdings[2].outputs[0], lotus.holdings[2].inputs[0].to_effect]
        left = FactorGroup(left)
        right = FactorGroup(oracle.holdings[2].outputs[0])
        context = next(left.likely_contexts(right))
        lotus_menu = lotus.holdings[2].generic_terms()[0]
        java_api = oracle.generic_terms()[0]
        assert context.get_factor(lotus_menu).compare_keys(java_api)

    def test_likely_context_from_factor_meaning(self, make_opinion_with_holding):
        lotus = make_opinion_with_holding["lotus_majority"]
        oracle = make_opinion_with_holding["oracle_majority"]
        left = lotus.holdings[2].outputs[0]
        right = oracle.holdings[2].outputs[0]
        likely = left._likely_context_from_meaning(right, context=ContextRegister())
        lotus_menu = lotus.holdings[2].generic_terms()[0]
        java_api = oracle.generic_terms()[0]
        assert likely.get_factor(lotus_menu).compare_keys(java_api)

    def test_union_one_generic_not_matched(self, make_opinion_with_holding):
        """
        Here, both FactorGroups have "fact that <> was a computer program".
        But they each have another generic that can't be matched:
        fact that <the Java API> was a literal element of <the Java language>
        and
        fact that <the Lotus menu command hierarchy> provided the means by
        which users controlled and operated <Lotus 1-2-3>

        Tests that Factors from "left" should be keys and Factors from "right" values.
        """
        lotus = make_opinion_with_holding["lotus_majority"]
        oracle = make_opinion_with_holding["oracle_majority"]
        left = FactorGroup(lotus.holdings[7].inputs[:2])
        right = FactorGroup(
            [oracle.holdings[3].outputs[0], oracle.holdings[3].inputs[0]]
        )
        new = left | right
        text = (
            "that <the Lotus menu command hierarchy> was a "
            "literal element of <Lotus 1-2-3>"
        )
        assert text in new[1].short_string


class TestChangeRegisters:
    def test_reverse_key_and_value_of_register(self):
        left = Entity("Siskel")
        right = Entity("Ebert")

        register = ContextRegister.from_lists([left], [right])

        assert len(register.keys()) == 1
        assert register.get("<Siskel>").compare_keys(Entity("Ebert"))

        new = register.reversed()
        assert new.get("<Ebert>").compare_keys(left)

    def test_factor_pairs(self):
        register = ContextRegister.from_lists(
            [Entity("apple"), Entity("lemon")], [Entity("pear"), Entity("orange")]
        )
        gen = register.factor_pairs()
        pair = next(gen)
        assert pair[0].name == "apple"
        assert pair[1].name == "pear"



"""
Tests of commands that appear in notebooks in
the notebooks/ directory
"""
from copy import deepcopy
import os

from dotenv import load_dotenv
import pytest

from authorityspoke.facts import Fact
from authorityspoke.io.downloads import CAPClient
from authorityspoke.io.readers import read_decision

from authorityspoke import Entity, Predicate, Comparison

load_dotenv()

CAP_API_KEY = os.getenv("CAP_API_KEY") or "wrong key"


class TestIntroduction:

    """
    Tests of commands from the "Introduction to AuthoritySpoke" notebook
    """

    client = CAPClient(api_token=CAP_API_KEY)

    @pytest.mark.vcr
    def test_download_case(self):
        oracle_download = self.client.fetch_cite(cite="750 F.3d 1339", full_case=True)
        oracle = read_decision(oracle_download)
        citations = [out_citation.cite for out_citation in oracle.cites_to]
        assert "527 F.3d 1318" in citations

    def test_oracle_20_holdings(self, make_opinion_with_holding):
        assert len(make_opinion_with_holding["oracle_majority"].holdings) == 20

    def test_replace_generic_factor(self, make_opinion_with_holding):
        lotus_majority = make_opinion_with_holding["lotus_majority"]

        seinfeld_holding = lotus_majority.holdings[0].new_context(
            terms_to_replace=[
                Entity("Borland International"),
                Entity("the Lotus menu command hierarchy"),
            ],
            changes=[Entity("Carol Publishing Group"), Entity("Seinfeld")],
        )

        assert lotus_majority.holdings[0] != seinfeld_holding
        assert lotus_majority.holdings[0].means(seinfeld_holding)

    def test_inferred_holdings_after_exclusive_holding(self, make_opinion_with_holding):
        """
        Test that when a holding is marked "exclusive" in the JSON input,
        that holding is added first to the Opinion's set of holdings, and
        any other inferred holdings, about the absence of the output from
        the original holding, are added later.
        """
        lotus_majority = make_opinion_with_holding["lotus_majority"]
        assert lotus_majority.holdings[0].outputs[0].absent is False
        assert lotus_majority.holdings[1].outputs[0].absent is True

    def test_change_rule_replacing_enactment(
        self, fake_usc_client, make_opinion_with_holding
    ):
        oracle = make_opinion_with_holding["oracle_majority"]

        works_of_authorship_passage = (
            "Copyright protection subsists, in accordance with this title, "
            + "in original works of authorship"
        )

        works_of_authorship_clause = fake_usc_client.read("/us/usc/t17/s102/a")
        works_of_authorship_clause.select(works_of_authorship_passage)
        holding_with_shorter_enactment = deepcopy(oracle.holdings[0])
        holding_with_shorter_enactment.set_enactments(works_of_authorship_clause)

        assert holding_with_shorter_enactment >= oracle.holdings[0]
        assert not oracle.holdings[0] >= holding_with_shorter_enactment

    def test_opinion_contradiction(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus_majority = make_opinion_with_holding["lotus_majority"]
        assert oracle.contradicts(lotus_majority)
        assert lotus_majority.contradicts(oracle)

    def test_opinion_explain_contradiction(self, make_opinion_with_holding):
        """
        The notebook now uses Decisions instead of this.
        """
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = lotus.holdings[6].explain_contradiction(oracle.holdings[10])
        assert (
            explanation.context["<the Lotus menu command hierarchy>"].name
            == "the Java API"
        )
        assert "<the Lotus menu command hierarchy> is like <the Java API>" in str(
            explanation
        )
        assert "Entity(name='the Java API'" in repr(explanation)

    def test_decision_explain_contradiction(self, make_decision_with_holding):
        oracle = make_decision_with_holding["oracle"]
        lotus = make_decision_with_holding["lotus"]
        explanation = lotus.explain_contradiction(oracle)
        lotus_like = "<the Lotus menu command hierarchy> is like <the Java API>"
        java_like = "<the Java API> is like <the Lotus menu command hierarchy>"
        assert lotus_like in str(explanation) or java_like in str(explanation)
        assert (
            "the Fact it was false that <the Lotus".lower() in str(explanation).lower()
        )

    def test_register_string(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = lotus.holdings[6].explain_contradiction(oracle.holdings[10])
        string = (
            "ContextRegister(<the Lotus menu command hierarchy> is like <the Java API>)"
        )
        assert str(explanation.context) == string

    def test_specific_holding_contradiction(self, make_opinion_with_holding):
        """
        Check the specific Holdings that should be causing a
        contradiction to be found between the Opinions.
        """
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        assert oracle.holdings[10].contradicts(lotus.holdings[6])

    def test_addition_some_to_some(self, make_opinion_with_holding):
        """
        Demonstrates that adding two SOME Holdings returns None,
        same as two SOME Rules.
        """

        oracle = make_opinion_with_holding["oracle_majority"]
        feist = make_opinion_with_holding["feist_majority"]
        listings_not_original = feist.holdings[10]
        original_not_copyrightable = oracle.holdings[0]
        assert listings_not_original + original_not_copyrightable is None

    def test_adding_holdings(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        listings_not_original = feist.holdings[10]
        unoriginal_not_copyrightable = feist.holdings[3]
        listings_not_copyrightable = (
            listings_not_original + unoriginal_not_copyrightable
        )
        not_copyrightable = unoriginal_not_copyrightable.outputs[0]
        assert listings_not_copyrightable.outputs[1].short_string == (
            "absence of the fact that <Rural's telephone"
            " listings> were copyrightable"
        )
        assert (
            "act that <Rural's telephone listings> were names, towns, "
            + "and telephone numbers of telephone subscribers"
        ) in listings_not_copyrightable.inputs[0].short_string

    def test_union_holdings_from_different_cases(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        feist = make_opinion_with_holding["feist_majority"]
        new = oracle.holdings[1] | feist.holdings[2]
        assert "it was false that <the Java API> was copyrightable" in str(new)
        assert "<the Java API> was an original work" in str(new)

    def test_only_one_explanation_for_contradiction(self, make_opinion_with_holding):
        lotus = make_opinion_with_holding["lotus_majority"]
        oracle = make_opinion_with_holding["oracle_majority"]

        gen = lotus.holdings[6].explanations_contradiction(oracle.holdings[10])
        first_explanation = next(gen)

        with pytest.raises(StopIteration):
            second_explanation = next(gen)


class TestTemplateStrings:
    """Tests from the notebook introducing template strings."""

    def test_no_line_break_in_fact_string(self):
        elaine = Entity("Elaine", generic=True)
        tax_rate_over_25 = Comparison(
            "${taxpayer}'s marginal income tax rate was", sign=">", expression=0.25
        )
        elaine_tax_rate = Fact(tax_rate_over_25, terms=elaine)
        assert "\n" not in str(elaine_tax_rate)

    def test_changing_order_of_concrete_terms_changes_meaning(self):
        ann = Entity("Ann", generic=False)
        bob = Entity("Bob", generic=False)
        parent_sentence = Predicate("$mother was ${child}'s parent")
        ann_parent = Fact(parent_sentence, terms=(ann, bob))
        bob_parent = Fact(parent_sentence, terms=(bob, ann))
        assert str(ann_parent).lower() == "the fact that Ann was Bob's parent".lower()
        assert str(bob_parent).lower() == "the fact that Bob was Ann's parent".lower()
        assert not ann_parent.means(bob_parent)



import pytest

from anchorpoint.textselectors import TextQuoteSelector

from nettlesome.entities import Entity

from authorityspoke.facts import Fact, Predicate
from authorityspoke.holdings import Holding, HoldingGroup
from authorityspoke.procedures import Procedure
from authorityspoke.rules import Rule
from authorityspoke.io import loaders, readers
from authorityspoke.io.fake_enactments import FakeClient
from authorityspoke.opinions import Opinion, FactorIndex


class TestOpinions:
    def test_opinion_features(self, make_decision):
        assert make_decision["watt"].court == "9th-cir"
        assert make_decision["watt"].citations[0].cite == "388 F.2d 853"

    def test_repr(self, make_opinion):
        assert "HAMLEY, Circuit Judge" in repr(make_opinion["watt_majority"])

    def test_repr_excludes_text(self, make_opinion):
        """
        reprs would be way too long if they could contain the
        full opinion text.
        """
        assert "text" not in repr(make_opinion["watt_majority"])

    def test_opinion_author(self, make_opinion):
        assert make_opinion["watt_majority"].author == "HAMLEY, Circuit Judge"
        assert make_opinion["brad_majority"].author == "BURKE, J."
        assert (
            make_opinion["brad_concurring-in-part-and-dissenting-in-part"].author
            == "TOBRINER, J."
        )

    def test_opinion_holding_list(self, make_opinion, real_holding):
        watt = make_opinion["watt_majority"]
        h3_specific = real_holding["h3"]
        watt.posit(h3_specific)
        assert h3_specific in watt.holdings

    def test_opinion_entity_list(
        self, make_opinion, real_holding, make_entity, make_evidence
    ):
        watt = make_opinion["watt_majority"]
        h = real_holding
        e = make_entity

        watt.posit(h["h1"], context=(e["motel"], e["watt"]))
        watt.posit(h["h2"], context=(e["trees"], e["motel"]))
        watt.posit(
            h["h3"],
            context=(
                make_evidence["generic"],
                e["motel"],
                e["watt"],
                e["trees"],
                e["tree_search"],
            ),
        )
        watt.posit(
            h["h4"], context=(e["trees"], e["tree_search"], e["motel"], e["watt"])
        )
        assert make_entity["watt"] in make_opinion["watt_majority"].generic_terms()

    def test_opinion_date(self, make_decision):
        """
        This no longer tests whether two Opinions from the same Decision
        have the same date, because the date field has been moved to the
        Decision class.
        """
        assert make_decision["watt"].date < make_decision["brad"].date

    def test_opinion_string(self, make_opinion):
        opinion = make_opinion["cardenas_majority"]
        assert str(opinion).lower() == "majority opinion by bird, c. j."


class TestOpinionText:
    def test_opinion_text(self, make_opinion):
        assert (
            "Feist responded that such efforts were economically "
            + "impractical and, in any event, unnecessary"
        ) in make_opinion["feist_majority"].text

    def test_opinion_text_anchor(self, make_opinion_with_holding):
        feist = make_opinion_with_holding["feist_majority"]
        assert any(
            "ideas" in anchor[0].exact for anchor in feist.factor_anchors.values()
        )

    def test_select_opinion_text_for_factor(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        anchor = oracle.holdings[0].anchors[0]
        selected = oracle.select_text(selector=anchor)
        assert selected == "must be “original” to qualify"

    def test_select_opinion_text_for_enactment(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        enactment_name = str(oracle.holdings[0].enactments[0])
        anchor = oracle.enactment_anchors[enactment_name][0]
        selected = oracle.select_text(selector=anchor)
        assert selected == "17 U.S.C. § 102(a)"

    def test_select_opinion_text_for_holding(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        holding = oracle.holdings[0]
        anchor = holding.anchors[0]
        selected = oracle.select_text(selector=anchor)
        assert selected == "must be “original” to qualify"

    def test_invalid_text_selector(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        anchor = TextQuoteSelector(exact="text not in opinion")
        with pytest.raises(ValueError):
            _ = oracle.select_text(selector=anchor)


class TestOpinionHoldings:
    def test_positing_non_rule_error(self, make_opinion, make_procedure):
        with pytest.raises(TypeError):
            make_opinion["watt_majority"].posit(make_procedure["c1"])

    def test_error_posit_with_no_rule_source(self, make_opinion):
        with pytest.raises(TypeError):
            make_opinion["watt_majority"].posit()

    def test_posit_rule(self, make_opinion, make_rule, make_holding):
        """
        "Positing" a Rule causes the Rule to be converted to a Holding first.
        So the Opinion implies the corresponding Holding.
        """

        watt = make_opinion["watt_majority"]
        watt.posit(make_rule["h1"])
        assert watt.implies(make_holding["h1"])

    def test_new_context_wrong_number_of_changes(self, make_opinion, make_holding):
        """
        The context here (a Factor outside an iterable) only changes the first
        generic factor of the Rule being posited, which may not be what the user
        expects.
        """
        brad = make_opinion["brad_majority"]
        with pytest.raises(ValueError):
            brad.posit(make_holding["h1"], context=Entity("House on Haunted Hill"))

    def test_new_context_naming_nonexistent_factor(self, make_opinion, make_holding):
        """
        The context here (a Factor outside an iterable) only changes the first
        generic factor of the Rule being posited, which may not be what the user
        expects.
        """
        brad = make_opinion["brad_majority"]
        brad.clear_holdings()
        with pytest.raises(ValueError):
            brad.posit(
                make_holding["h1"],
                context=(Entity("House on Haunted Hill"), "nonexistent factor"),
            )

    def test_new_context_creates_equal_rule(self, make_opinion, make_response):
        watt = make_opinion["watt_majority"]
        brad = make_opinion["brad_majority"]
        client = FakeClient(responses=make_response)

        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.json")
        watt.posit(readers.read_holdings(watt_raw, client=client))

        brad.clear_holdings()
        brad_raw = loaders.load_holdings("holding_brad.json")
        brad.posit(readers.read_holdings(brad_raw, client=client))

        context_pairs = {
            "proof of Bradley's guilt": "proof of Wattenburg's guilt",
            "Bradley": "Wattenburg",
            "officers' search of the yard": "officers' search of the stockpile",
            "Bradley's marijuana patch": "the stockpile of trees",
        }
        watt.posit(brad.holdings[0], context_pairs)
        assert watt.holdings[-1].means(brad.holdings[0])

    def test_getting_factors_from_opinion(self, make_opinion, make_response):
        client = FakeClient(responses=make_response)

        watt = make_opinion["watt_majority"]
        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.json")
        holdings_to_posit = readers.read_holdings(watt_raw, client=client)
        watt.posit(holdings_to_posit)
        factors = watt.factors_by_name()
        assert "proof of Wattenburg's guilt" in factors.keys()

    def test_new_context_inferring_factors_to_change(self, make_opinion, make_response):
        """
        This changes watt's holdings; may break tests below.
        """
        watt = make_opinion["watt_majority"]
        brad = make_opinion["brad_majority"]

        client = FakeClient(responses=make_response)

        watt.clear_holdings()
        watt_raw = loaders.load_holdings("holding_watt.json")
        watt.posit(readers.read_holdings(watt_raw, client=client))

        brad.clear_holdings()
        brad_raw = loaders.load_holdings("holding_brad.json")
        brad.posit(readers.read_holdings(brad_raw, client=client))

        context_items = [
            "proof of Wattenburg's guilt",
            "<Wattenburg>",
            "<officers' search of the stockpile>",
            "<Hideaway Lodge>",
            "<the stockpile of trees>",
        ]
        watt.posit(brad.holdings[0], context=context_items)
        assert watt.holdings[-1].means(brad.holdings[0])

    def test_getting_factors_from_new_holding(self, make_opinion):
        watt = make_opinion["watt_majority"]
        elephants = Fact("$animal was an elephant", terms=Entity("the elephant"))
        mouseholes = Fact(
            Predicate("$animal hides in mouseholes", truth=False),
            terms=Entity("the elephant"),
        )
        procedure = Procedure(inputs=elephants, outputs=mouseholes)
        rule = Rule(procedure=procedure)
        holding = Holding(rule=rule)
        watt.posit(holding)
        factors = watt.factors_by_name()
        factor = factors["the fact that <the elephant> was an elephant"]
        assert factor.terms[0].name == "the elephant"


class TestOpinionFactors:
    def test_only_one_factor_with_same_content(self, make_opinion_with_holding):
        """
        Tests that a particular Factor appears only once, and that all
        three of the text anchors for that Factor appear in the value
        for the Factor in Opinion.factors.
        """

        oracle = make_opinion_with_holding["oracle_majority"]
        scenes_a_faire = [
            factor
            for factor in oracle.factors()
            if isinstance(factor, Fact)
            and factor.short_string
            == "the fact that <the Java API> was a scene a faire"
        ]
        assert len(scenes_a_faire) == 1  # 1 Factor

    def test_insert_duplicate_anchor_in_factor_index(self):
        api = Entity(name="the Java API", generic=True, plural=False)
        anchor = TextQuoteSelector(
            exact="it possesses at least some minimal degree of creativity."
        )
        fact = Fact(
            predicate=Predicate(
                "$product possessed at least some minimal degree of creativity"
            ),
            terms=[api],
            anchors=[anchor],
        )
        name = "the Java API possessed at least some minimal degree of creativity"

        factor_index = FactorIndex({name: fact})
        factor_index.insert(key=name, value=fact)
        assert len(factor_index[name].anchors) == 1

    def test_get_factor_from_opinion(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        company = oracle.get_factor_by_name("the Java API")
        assert isinstance(company, Entity)

    def test_factors_by_name(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        factors = oracle.factors_by_name()
        factor = factors[
            "the fact it was false that <the Java API> was an original work"
        ]
        assert factor.terms[0].name == "the Java API"


class TestImplication:
    def test_no_implication(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        brad = make_opinion_with_holding["brad_majority"]
        assert not watt >= brad
        assert not watt.implies(brad.holdings)

    def test_posit_list_of_holdings_and_imply(self, make_opinion, make_response):
        watt = make_opinion["watt_majority"]
        brad = make_opinion["brad_majority"]
        client = FakeClient(responses=make_response)
        some_rules_raw = loaders.load_holdings(filename="holding_watt.json")
        some_rules = readers.read_holdings(some_rules_raw, client=client)
        for case in (watt, brad):
            case.clear_holdings()
            case.posit(some_rules[:3])
        watt.posit(some_rules[3])
        assert watt > brad
        assert not brad >= watt

    def test_error_to_compare_to_str(self, make_opinion):
        watt = make_opinion["watt_majority"]
        with pytest.raises(TypeError):
            watt.implies("this")

    def test_error_to_check_contradiction_of_str(self, make_opinion):
        watt = make_opinion["watt_majority"]
        with pytest.raises(TypeError):
            watt.contradicts("this")

    def test_opinion_implies_None(self, make_opinion, make_holding):
        watt = make_opinion["watt_majority"]
        assert watt.implies(None)

    def test_opinion_implies_holding(self, make_opinion, make_holding):
        watt = make_opinion["watt_majority"]
        watt.posit(make_holding["h2_invalid_undecided"])
        assert watt >= make_holding["h2_undecided"]
        assert watt > make_holding["h2_undecided"]

    def test_opinion_does_not_imply_holding(self, make_opinion, make_holding):
        watt = make_opinion["watt_majority"]
        watt.clear_holdings()
        watt.posit(make_holding["h2_irrelevant_inputs_undecided"])
        assert not watt >= make_holding["h2_undecided"]
        assert not watt > make_holding["h2_undecided"]

    def test_opinion_implied_by_rule(self, make_opinion, make_holding, make_rule):
        watt = make_opinion["oracle_majority"]
        watt.clear_holdings()
        watt.posit(make_holding["h2"])
        assert watt.implied_by(make_rule["h2_despite_due_process"])
        assert not watt.implies(make_rule["h2_despite_due_process"])

    def test_opinion_implies_holding_group(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        holdings = watt.holdings[:2]
        assert isinstance(holdings, HoldingGroup)
        assert watt.implies(holdings)
        explanation = watt.explain_implication(holdings)
        assert len(explanation.reasons) == 2

    def test_opinion_does_not_imply_holding_group(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        holdings = make_opinion_with_holding["oracle_majority"].holdings
        assert isinstance(holdings, HoldingGroup)
        assert not watt.implies(holdings)
        explanation = watt.explain_implication(holdings)
        assert explanation is None

    def test_opinion_implied_by_decision(
        self, make_holding, make_decision_with_holding
    ):
        watt = make_decision_with_holding["watt"]
        opinion = Opinion()
        opinion.posit(watt.holdings[0])
        assert opinion.implied_by(watt)


class TestContradiction:
    def test_opinion_contradicts_opinion(self, make_opinion_with_holding):
        """Return the only contradictory pair of Holdings between these two Opinions."""
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]

        left = Opinion()
        left.posit(lotus.holdings[6])
        right = Opinion()
        right.posit(oracle.holdings[10])
        explanation = left.explain_contradiction(right)
        assert len(explanation.reasons) == 1
        assert isinstance(explanation.reasons[0].left, Holding)
        assert isinstance(explanation.reasons[0].right, Holding)

    def test_contradiction_of_holding(
        self, make_opinion_with_holding, e_search_clause, make_holding
    ):
        assert make_opinion_with_holding["watt_majority"].contradicts(
            make_holding["h2_output_false_ALL_MUST"] + e_search_clause
        )

    def test_explain_opinion_contradicting_holding(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = oracle.explain_contradiction(lotus.holdings[6])
        assert (
            "<the java api> is like <the lotus menu command hierarchy>"
            in str(explanation).lower()
        )

    def test_explain_opinion_contradicting_rule(self, make_opinion_with_holding):
        oracle = make_opinion_with_holding["oracle_majority"]
        lotus = make_opinion_with_holding["lotus_majority"]
        explanation = oracle.explain_contradiction(lotus.holdings[6].rule)
        expected = "<the java api> is like <the lotus menu command hierarchy>"
        assert expected in str(explanation).lower()

    def test_contradiction_of_decision(
        self, make_opinion_with_holding, make_decision_with_holding
    ):
        assert make_opinion_with_holding["oracle_majority"].contradicts(
            make_decision_with_holding["lotus"]
        )

    def test_explain_opinion_contradicting_decision(
        self, make_opinion_with_holding, make_decision_with_holding
    ):
        oracle_majority = make_opinion_with_holding["oracle_majority"]
        lotus = make_decision_with_holding["lotus"]
        explanation = oracle_majority.explain_contradiction(lotus)
        assert "contradicts" in str(explanation).lower()

    def test_no_explanation_of_contradiction(self, make_opinion_with_holding):
        watt = make_opinion_with_holding["watt_majority"]
        holdings = watt.holdings[:2]
        assert isinstance(holdings, HoldingGroup)
        assert watt.implies(holdings)
        explanation = watt.explain_contradiction(holdings)
        assert explanation is None

    def test_error_contradiction_with_procedure(self, make_opinion, make_procedure):
        with pytest.raises(TypeError):
            make_opinion["watt_majority"].contradicts(make_procedure["c1"])



import operator

import pytest

from nettlesome.terms import ContextRegister, DuplicateTermError
from nettlesome.terms import Explanation, TermSequence, means
from nettlesome.entities import Entity
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison, Q_

from authorityspoke.facts import Fact, build_fact


class TestFacts:
    def test_default_terms_for_fact(self, make_entity, make_predicate, watt_mentioned):
        e = make_entity
        f1 = build_fact(make_predicate["p1"], case_factors=watt_mentioned)
        assert f1.terms == (e["motel"],)

    def test_build_fact(self, make_predicate, watt_mentioned):
        """
        Check that terms is created as a (hashable) tuple, not list
        """
        shooting = build_fact(
            make_predicate["p_shooting"],
            (2, 3),
            case_factors=watt_mentioned,
            standard_of_proof="preponderance of evidence",
        )
        assert isinstance(shooting.terms, tuple)

    def test_terms_from_case_factor_indices(
        self, make_entity, make_predicate, watt_mentioned
    ):
        """
        If you pass in integers instead of Factor objects to fill the blanks
        in the Predicate (which was the only way to do things in the first
        version of the Fact class's __init__ method), then the integers
        you pass in should be used as indices to select Factor objects
        from case_factors.
        """

        e = make_entity

        f2 = build_fact(
            make_predicate["p2"], indices=(1, 0), case_factors=watt_mentioned
        )
        assert f2.terms == (e["watt"], e["motel"])

    def test_correct_factors_from_indices_in_build_fact(
        self, make_entity, make_predicate, watt_mentioned
    ):
        e = make_entity
        f2 = build_fact(
            make_predicate["p2"],
            indices=(1, 2),
            case_factors=watt_mentioned,
        )
        assert f2.terms == (e["watt"], e["trees"])

    def test_wrong_type_in_terms_in_init(
        self, make_entity, make_predicate, watt_mentioned
    ):
        e = make_entity
        with pytest.raises(TypeError):
            f2 = build_fact(
                make_predicate["p1"],
                indices=("nonsense"),
                case_factors=watt_mentioned,
            )

    def test_invalid_index_for_case_factors_in_init(self, make_predicate, make_entity):
        with pytest.raises(IndexError):
            _ = build_fact(
                make_predicate["p1"],
                indices=2,
                case_factors=make_entity["watt"],
            )

    def test_convert_int_terms_to_tuple(self, make_predicate, watt_mentioned):
        f = build_fact(make_predicate["p_irrelevant_1"], 3, case_factors=watt_mentioned)
        assert f.terms == (watt_mentioned[3],)

    def test_string_representation_of_factor(self, watt_factor):
        assert "<Hideaway Lodge> was a motel" in str(watt_factor["f1"])
        assert "absence of the fact" in str(watt_factor["f3_absent"]).lower()

    def test_string_no_truth_value(self, watt_factor):
        factor = watt_factor["f2_no_truth"]
        assert "whether" in str(factor)

    def test_repeating_entity_string(self, make_predicate):
        """I'm not convinced that a model of a Fact ever needs to include
        multiple references to the same Entity just because the name of the
        Entity appears more than once in the Predicate."""
        with pytest.raises(DuplicateTermError):
            Fact(
                make_predicate["p_three_entities"],
                terms=[Entity("Al"), Entity("Bob"), Entity("Al")],
            )

    def test_string_representation_with_concrete_entities(self, watt_factor):
        """
        "Hideaway Lodge" is still a string representation of an Entity
        object, but it's not in angle brackets because it can't be
        replaced by another Entity object without changing the meaning
        of the Fact.
        """
        assert "Hideaway Lodge was a motel" in str(watt_factor["f1_specific"])

    def test_string_for_fact_with_identical_terms(self):
        devon = Entity("Devon", generic=True)
        elaine = Entity("Elaine", generic=True)
        opened_account = Fact(
            Predicate("$applicant opened a bank account for $applicant and $cosigner"),
            terms=(devon, elaine),
        )
        assert "<Devon> opened a bank account for <Devon> and <Elaine>" in str(
            opened_account
        )

    def test_str_with_concrete_context(self, make_opinion_with_holding):
        holding = list(make_opinion_with_holding["cardenas_majority"].holdings)[1]
        longer_str = holding.inputs[0].str_with_concrete_context
        assert "the exhibit in the form testimony" in longer_str.lower()
        assert "the exhibit in the form testimony" not in str(holding.inputs[0]).lower()

    def test_complex_fact_no_line_break_in_predicate(self, make_opinion_with_holding):
        """
        Tests that the string representation of this Holding's only input
        Fact does not contain indented new lines, except in the "SPECIFIC
        CONTEXT" part, if present.

        The representation of the Exhibit mentioned in the Fact should
        not introduce any indented lines inside the Fact's string.
        """
        holding = list(make_opinion_with_holding["cardenas_majority"].holdings)[1]
        fact_text = str(holding.inputs[0])
        if "SPECIFIC CONTEXT" in fact_text:
            fact_text = fact_text.split("SPECIFIC CONTEXT")[0].strip()
        assert "\n  " not in fact_text

    def test_new_context_replace_fact(self, make_entity, watt_factor):
        changes = ContextRegister.from_lists(
            [make_entity["watt"], watt_factor["f2"]],
            [Entity("Darth Vader"), watt_factor["f10"]],
        )
        assert "was within the curtilage of <Hideaway Lodge>" in (
            watt_factor["f2"].new_context(changes).short_string
        )

    def test_get_factor_from_recursive_search(self, make_opinion_with_holding):
        holding_list = list(make_opinion_with_holding["cardenas_majority"].holdings)
        factor_list = list(holding_list[0].recursive_terms.values())
        assert any(
            factor.compare_keys(Entity("parole officer"))
            and factor.name == "parole officer"
            for factor in factor_list
        )

    def test_no_duplicate_comma_in_evidence(self, make_opinion_with_holding):
        holdings = list(make_opinion_with_holding["cardenas_majority"].holdings)
        assert "addicted to heroin, in showing the fact" in str(holdings[0].outputs[0])

    def test_new_context_from_factor(self, watt_factor):
        different = watt_factor["f1"].new_context(
            Entity("Great Northern", generic=False)
        )
        assert "Great Northern was a motel" in str(different)

    def test_new_concrete_context(self, make_entity, watt_factor):
        register = ContextRegister.from_lists(
            to_replace=[make_entity["watt"], make_entity["motel"]],
            replacements=[Entity("Darth Vader"), Entity("Death Star")],
        )
        different = watt_factor["f2"].new_context(register)
        assert "<Darth Vader> operated" in str(different)

    def test_type_of_terms(self, watt_factor):
        assert isinstance(watt_factor["f1"].terms, TermSequence)

    def test_concrete_to_abstract(self, make_entity, make_predicate):
        motel = make_entity["motel_specific"]
        d = make_entity["watt"]
        fact = Fact(predicate=make_predicate["p2"], terms=(d, motel))
        assert "<Wattenburg> operated and lived at Hideaway Lodge" in str(fact)
        assert "<Wattenburg> operated and lived at Hideaway Lodge>" in str(
            fact.make_generic()
        )

    def test_entity_slots_as_length_of_factor(self, watt_factor):
        assert len(watt_factor["f1"].predicate) == 1
        assert len(watt_factor["f1"]) == 1

    def test_predicate_with_entities(self, make_entity, watt_factor):
        assert "<Hideaway Lodge> was a motel" in str(watt_factor["f1"])

    def test_factor_terms_do_not_match_predicate(self, make_predicate, watt_mentioned):
        """
        make_predicate["p1"] has only one slot for context factors, but
        this tells it to look for three.
        """
        with pytest.raises(ValueError):
            _ = build_fact(make_predicate["p1"], (0, 1, 2), case_factors=watt_mentioned)

    def test_reciprocal_with_wrong_number_of_entities(self, make_entity, watt_factor):
        with pytest.raises(ValueError):
            watt_factor["f1"].predicate._content_with_terms(
                (make_entity["motel"], make_entity["watt"])
            )

    def test_entity_and_human_in_predicate(self, make_entity, watt_factor):
        assert "<Wattenburg> operated and lived at <Hideaway Lodge>" in watt_factor[
            "f2"
        ].predicate._content_with_terms((make_entity["watt"], make_entity["motel"]))

    def test_standard_of_proof_must_be_listed(self, make_predicate, watt_mentioned):
        with pytest.raises(ValueError):
            _ = build_fact(
                make_predicate["p2"],
                case_factors=watt_mentioned,
                standard_of_proof="probably so",
            )

    def test_standard_of_proof_in_str(self, watt_factor):
        factor = watt_factor["f2_preponderance_of_evidence"]
        assert factor.standard_of_proof in factor.short_string

    def test_case_factors_deleted_from_fact(self, watt_factor):
        """This attribute should have been deleted during Fact.__post_init__"""
        predicate = Predicate("some things happened")
        factor = build_fact(predicate)
        assert not hasattr(factor, "case_factors")

    def test_repeated_placeholder_in_fact(self, make_opinion_with_holding):
        holding = make_opinion_with_holding["lotus_majority"].holdings[9]
        fact = holding.inputs[1]
        assert fact.short_string == (
            "the fact it was false that the precise formulation "
            "of <Lotus 1-2-3>'s code was necessary for <Lotus 1-2-3> to work"
        )
        assert len(fact.terms) == 1


class TestSameMeaning:
    def test_equality_factor_from_same_predicate(self, watt_factor):
        assert watt_factor["f1"].means(watt_factor["f1b"])

    def test_equality_factor_from_equal_predicate(self, watt_factor):
        assert watt_factor["f1"].means(watt_factor["f1c"])

    def test_equality_because_factors_are_generic_entities(self, watt_factor):
        assert watt_factor["f1"].means(watt_factor["f1_different_entity"])

    def test_unequal_because_a_factor_is_not_generic(self, watt_factor):
        assert not watt_factor["f9_swap_entities_4"].means(watt_factor["f9"])

    def test_generic_terms_equal(self, watt_factor):
        assert watt_factor["f2_generic"].means(watt_factor["f2_false_generic"])
        assert watt_factor["f2_generic"].means(watt_factor["f3_generic"])

    def test_equal_referencing_diffent_generic_terms(self, make_factor):
        assert make_factor["f_murder"].means(make_factor["f_murder_craig"])

    def test_generic_and_specific_factors_unequal(self, watt_factor):
        assert not watt_factor["f2"].means(watt_factor["f2_generic"])

    def test_factor_reciprocal_unequal(self, watt_factor):
        assert not watt_factor["f2"].means(watt_factor["f2_reflexive"])

    def test_factor_different_predicate_truth_unequal(self, watt_factor):
        assert not watt_factor["f7"].means(watt_factor["f7_opposite"])

    def test_unequal_because_one_factor_is_absent(self, watt_factor):
        assert not watt_factor["f8"].means(watt_factor["f8_absent"])

    def test_copies_of_identical_factor(self, make_factor):
        """
        Even if the two factors have different entity markers in self.terms,
        I expect them to evaluate equal because the choice of entity markers is
        arbitrary.
        """
        f = make_factor
        assert f["f_irrelevant_3"].means(f["f_irrelevant_3"])
        assert f["f_irrelevant_3"].means(f["f_irrelevant_3_new_context"])

    def test_equal_with_different_generic_subfactors(self, make_complex_fact):
        assert make_complex_fact["f_relevant_murder"].means(
            make_complex_fact["f_relevant_murder_craig"]
        )

    def test_reciprocal_context_register(self, watt_factor):
        """
        This test describes two objects with the same meaning that have been
        made in two different ways, each with a different id and repr.
        """
        assert watt_factor["f7"].means(watt_factor["f7_swap_entities"])

    def test_interchangeable_concrete_terms(self):
        """Detect that placeholders differing only by a final digit are interchangeable."""
        ann = Entity("Ann", generic=False)
        bob = Entity("Bob", generic=False)

        ann_and_bob_were_family = Fact(
            Predicate("$relative1 and $relative2 both were members of the same family"),
            terms=(ann, bob),
        )
        bob_and_ann_were_family = Fact(
            Predicate("$relative1 and $relative2 both were members of the same family"),
            terms=(bob, ann),
        )

        assert ann_and_bob_were_family.means(bob_and_ann_were_family)

    def test_unequal_to_enactment(self, watt_factor, e_copyright):
        assert not watt_factor["f1"].means(e_copyright)
        with pytest.raises(TypeError):
            e_copyright.means(watt_factor["f1"])

    def test_standard_of_proof_inequality(self, watt_factor):

        f = watt_factor
        assert not f["f2_clear_and_convincing"].means(f["f2_preponderance_of_evidence"])
        assert not f["f2_clear_and_convincing"].means(f["f2"])

    def test_means_despite_plural(self):
        directory = Entity("Rural's telephone directory", plural=False)
        listings = Entity("Rural's telephone listings", plural=True)
        directory_original = Fact(Predicate("$thing was original"), terms=directory)
        listings_original = Fact(Predicate("$thing were original"), terms=listings)
        assert directory_original.means(listings_original)

    def test_same_meaning_no_terms(self, make_factor):
        assert make_factor["f_no_context"].means(make_factor["f_no_context"])


class TestImplication:
    def test_fact_implies_none(self, watt_factor):
        assert watt_factor["f1"].implies(None)

    def test_no_implication_of_rule(self, watt_factor, make_rule):
        assert not watt_factor["f1"].implies(make_rule["h1"])

    def test_fact_does_not_imply_holding(self, watt_factor, make_holding):
        assert not watt_factor["f1"].implies(make_holding["h1"])

    def test_specific_factor_implies_generic(self, watt_factor):
        assert watt_factor["f2"] > watt_factor["f2_generic"]
        assert not watt_factor["f2_generic"] > watt_factor["f2"]

    def test_specific_factor_implies_generic_explain(self, watt_factor):
        answer = watt_factor["f2"].explain_implication(watt_factor["f2_generic"])
        assert (
            str(watt_factor["f2"]),
            watt_factor["f2_generic"],
        ) in answer.context.items()

    def test_specific_implies_generic_form_of_another_fact(self, watt_factor):
        assert watt_factor["f2"] > watt_factor["f3_generic"]

    def test_specific_fact_does_not_imply_generic_entity(
        self, make_entity, watt_factor
    ):
        assert not watt_factor["f2"] > make_entity["motel"]

    def test_factor_does_not_imply_predicate(self, make_predicate, watt_factor):
        with pytest.raises(TypeError):
            assert not watt_factor["f8_meters"] > make_predicate["p8"]

    def test_factor_implies_because_of_quantity(self, watt_factor):
        assert watt_factor["f8_meters"] > watt_factor["f8"]
        assert watt_factor["f8_higher_int"] > watt_factor["f8_float"]
        assert watt_factor["f8_int"] < watt_factor["f8_higher_int"]

    def test_factor_implies_no_truth_value(self, watt_factor):
        assert watt_factor["f2"] > watt_factor["f2_no_truth"]
        assert not watt_factor["f2_no_truth"] > watt_factor["f2"]

    def test_comparison_implies_no_truth_value(self, watt_factor):
        assert watt_factor["f8"] > watt_factor["f8_no_truth"]
        assert not watt_factor["f8_no_truth"] > watt_factor["f8"]

    def test_implication_standard_of_proof(self, make_factor):
        assert (
            not make_factor["f_shooting_craig_poe"]
            > make_factor["f_shooting_craig_brd"]
        )
        assert make_factor["f_shooting_craig_brd"] > make_factor["f_shooting_craig_poe"]

    def test_factor_implies_because_of_exact_quantity(self, watt_factor):
        assert watt_factor["f8_exact"] > watt_factor["f7"]
        assert watt_factor["f8_exact"] >= watt_factor["f8"]

    def test_no_implication_pint_quantity_and_int(self, watt_factor):
        assert not watt_factor["f8"] > watt_factor["f8_int"]
        assert not watt_factor["f8"] < watt_factor["f8_int"]

    def test_absent_factor_implies_absent_factor_with_lesser_quantity(
        self, watt_factor
    ):
        assert watt_factor["f9_absent_miles"] > watt_factor["f9_absent"]

    def test_equal_factors_not_gt(self, watt_factor):
        f = watt_factor
        assert f["f7"] >= f["f7"]
        assert f["f7"] <= f["f7"]
        assert not f["f7"] > f["f7"]

    def test_standard_of_proof_comparison(self, watt_factor):
        f = watt_factor
        assert f["f2_clear_and_convincing"].implies(f["f2_preponderance_of_evidence"])
        assert f["f2_beyond_reasonable_doubt"] >= f["f2_clear_and_convincing"]

    def test_no_implication_between_factors_with_and_without_standards(
        self, watt_factor
    ):
        f = watt_factor
        assert not f["f2_clear_and_convincing"] > f["f2"]
        assert not f["f2"] > f["f2_preponderance_of_evidence"]

    def test_implication_complex(self, make_complex_fact):
        assert (
            make_complex_fact["f_relevant_murder"]
            > make_complex_fact["f_relevant_murder_whether"]
        )

    def test_context_register_text(self, make_context_register):
        assert str(make_context_register) == (
            "ContextRegister(<Alice> is like <Craig>, and <Bob> is like <Dan>)"
        )

    def test_implication_complex_explain(
        self, make_complex_fact, make_context_register
    ):
        complex_true = make_complex_fact["f_relevant_murder"]
        complex_whether = make_complex_fact["f_relevant_murder_whether"].new_context(
            make_context_register
        )
        explanation = complex_true.explain_implication(complex_whether)
        assert explanation.context[Entity("Alice").key].compare_keys(Entity("Craig"))

    def test_implication_explain_keys_only_from_left(
        self, make_complex_fact, make_context_register
    ):
        """
        Check that when implies provides a ContextRegister as an "explanation",
        it uses elements only from the left as keys and from the right as values.
        """
        complex_true = make_complex_fact["f_relevant_murder"]
        complex_whether = make_complex_fact["f_relevant_murder_whether"]
        new = complex_whether.new_context(make_context_register)
        explanations = list(complex_true.explanations_implication(new))
        explanation = explanations.pop()
        assert explanation.context.get("<Craig>") != Entity("Alice")
        assert explanation.context.get("<Alice>").compare_keys(Entity("Craig"))

    def test_context_registers_for_complex_comparison(self, make_complex_fact):
        gen = make_complex_fact["f_relevant_murder_nested_swap"]._context_registers(
            make_complex_fact["f_relevant_murder"], operator.ge
        )
        register = next(gen)
        assert register.matches.get("<Alice>").compare_keys(Entity("Bob"))

    def test_no_implication_complex(self, make_complex_fact):
        left = make_complex_fact["f_relevant_murder"]
        right = make_complex_fact["f_relevant_murder_alice_craig"]
        assert not left >= right
        assert left.explain_implication(right) is None

    def test_implied_by(self, make_complex_fact):
        assert make_complex_fact["f_relevant_murder_whether"].implied_by(
            make_complex_fact["f_relevant_murder"]
        )

    def test_explanation_implied_by(self, make_complex_fact):
        explanation = make_complex_fact["f_relevant_murder_whether"].explain_implied_by(
            make_complex_fact["f_relevant_murder"]
        )
        assert explanation

    def test_explain_not_implied_by(self, make_complex_fact):
        left = make_complex_fact["f_relevant_murder"]
        right = make_complex_fact["f_relevant_murder_whether"]
        assert left.explain_implied_by(right) is None

    def test_not_implied_by_none(self, make_complex_fact):
        left = make_complex_fact["f_relevant_murder"]
        assert not left.implied_by(None)


class TestContradiction:
    def test_factor_different_predicate_truth_contradicts(self, watt_factor):
        assert watt_factor["f7"].contradicts(watt_factor["f7_opposite"])
        assert watt_factor["f7_opposite"].contradicts(watt_factor["f7"])

    def test_same_predicate_true_vs_false(self, watt_factor):
        assert watt_factor["f10"].contradicts(watt_factor["f10_false"])
        assert watt_factor["f10"].truth != watt_factor["f10_false"].truth

    def test_factor_does_not_contradict_predicate(self, make_predicate, watt_factor):
        with pytest.raises(TypeError):
            _ = watt_factor["f7"].contradicts(make_predicate["p7_true"])

    def test_factor_contradiction_absent_predicate(self, watt_factor):
        assert watt_factor["f3"].contradicts(watt_factor["f3_absent"])
        assert watt_factor["f3_absent"].contradicts(watt_factor["f3"])

    def test_absences_of_contradictory_facts_consistent(self, watt_factor):
        assert not watt_factor["f8_absent"].contradicts(watt_factor["f8_less_absent"])

    def test_factor_no_contradiction_no_truth_value(self, watt_factor):
        assert not watt_factor["f2"].contradicts(watt_factor["f2_no_truth"])
        assert not watt_factor["f2_no_truth"].contradicts(watt_factor["f2_false"])

    def test_absent_factor_contradicts_broader_quantity_statement(self, watt_factor):
        assert watt_factor["f8_absent"].contradicts(watt_factor["f8_meters"])
        assert watt_factor["f8_meters"].contradicts(watt_factor["f8_absent"])

    def test_less_specific_absent_contradicts_more_specific(self, watt_factor):
        assert watt_factor["f9_absent_miles"].contradicts(watt_factor["f9"])
        assert watt_factor["f9"].contradicts(watt_factor["f9_absent_miles"])

    def test_no_contradiction_with_more_specific_absent(self, watt_factor):
        assert not watt_factor["f9_absent"].contradicts(watt_factor["f9_miles"])
        assert not watt_factor["f9_miles"].contradicts(watt_factor["f9_absent"])

    def test_contradiction_complex(self, make_complex_fact):
        assert make_complex_fact["f_irrelevant_murder"].contradicts(
            make_complex_fact["f_relevant_murder_craig"]
        )

    def test_no_contradiction_complex(self, make_complex_fact):
        assert not make_complex_fact["f_irrelevant_murder"].contradicts(
            make_complex_fact["f_relevant_murder_alice_craig"]
        )

    def test_no_contradiction_of_None(self, watt_factor):
        assert not watt_factor["f1"].contradicts(None)

    def test_contradicts_if_present_both_present(self, watt_factor):
        """
        Test a helper function that checks whether there would
        be a contradiction if neither Factor was "absent".
        """
        assert watt_factor["f2"]._contradicts_if_present(
            watt_factor["f2_false"], explanation=Explanation.from_context()
        )

    def test_contradicts_if_present_one_absent(self, watt_factor):
        assert watt_factor["f2"]._contradicts_if_present(
            watt_factor["f2_false_absent"], explanation=Explanation.from_context()
        )

    def test_false_does_not_contradict_absent(self):
        absent_fact = Fact(
            predicate=Predicate(
                content="${rural_s_telephone_directory} was copyrightable", truth=True
            ),
            terms=(Entity(name="Rural's telephone directory")),
            absent=True,
        )
        false_fact = Fact(
            predicate=Predicate(
                content="${the_java_api} was copyrightable", truth=False
            ),
            terms=(Entity(name="the Java API", generic=True, plural=False)),
            absent=False,
        )
        assert not false_fact.contradicts(absent_fact)
        assert not absent_fact.contradicts(false_fact)

    def test_inconsistent_statements_about_different_entities(self):
        """
        Alice and Bob are both generics. So it's possible to reach a
        contradiction if you assume they correspond to one another.
        """
        p_small_weight = Comparison(
            "the amount of gold $person possessed was",
            sign="<",
            expression=Q_("1 gram"),
        )
        p_large_weight = Comparison(
            "the amount of gold $person possessed was",
            sign=">=",
            expression=Q_("100 kilograms"),
        )
        alice = Entity("Alice")
        bob = Entity("Bob")
        alice_rich = Fact(p_large_weight, terms=alice)
        bob_poor = Fact(p_small_weight, terms=bob)
        assert alice_rich.contradicts(bob_poor)

    def test_inconsistent_statements_about_corresponding_entities(self):
        """
        Even though Alice and Bob are both generics, it's known that
        Alice in the first context corresponds with Alice in the second.
        So there's no contradiction.
        """
        p_small_weight = Comparison(
            "the amount of gold $person possessed was",
            sign="<",
            expression=Q_("1 gram"),
        )
        p_large_weight = Comparison(
            "the amount of gold $person possessed was",
            sign=">=",
            expression=Q_("100 kilograms"),
        )
        alice = Entity("Alice")
        bob = Entity("Bob")
        alice_rich = Fact(p_large_weight, terms=alice)
        bob_poor = Fact(p_small_weight, terms=bob)
        register = ContextRegister()
        register.insert_pair(alice, alice)
        assert not alice_rich.contradicts(bob_poor, context=register)

    def test_copy_with_foreign_context(self, watt_mentioned, watt_factor):
        w = watt_mentioned
        assert (
            watt_factor["f1"]
            .new_context(ContextRegister.from_lists([w[0]], [w[2]]))
            .means(watt_factor["f1_different_entity"])
        )

    def test_check_entity_consistency_true(self, make_entity, make_factor):
        left = make_factor["f_irrelevant_3"]
        right = make_factor["f_irrelevant_3_new_context"]
        e = make_entity
        easy_register = ContextRegister.from_lists([e["dan"]], [e["craig"]])
        easy_update = left.update_context_register(
            right, easy_register, comparison=means
        )
        harder_register = ContextRegister.from_lists(
            to_replace=[e["alice"], e["bob"], e["craig"], e["dan"], e["circus"]],
            replacements=[e["bob"], e["alice"], e["dan"], e["craig"], e["circus"]],
        )
        harder_update = left.update_context_register(
            right,
            context=harder_register,
            comparison=means,
        )
        assert any(register is not None for register in easy_update)
        assert any(register is not None for register in harder_update)

    def test_check_entity_consistency_false(self, make_entity, make_factor):
        context = ContextRegister()
        context.insert_pair(make_entity["circus"], make_entity["alice"])
        update = make_factor["f_irrelevant_3"].update_context_register(
            make_factor["f_irrelevant_3_new_context"], comparison=means, context=context
        )
        assert not any(register is not None for register in update)

    def test_entity_consistency_identity_not_equality(self, make_entity, make_factor):

        register = ContextRegister()
        register.insert_pair(make_entity["dan"], make_entity["dan"])
        update = make_factor["f_irrelevant_3"].update_context_register(
            make_factor["f_irrelevant_3_new_context"],
            context=register,
            comparison=means,
        )
        assert not any(register is not None for register in update)

    def test_check_entity_consistency_type_error(
        self, make_entity, make_factor, make_predicate
    ):
        """
        There would be no TypeError if it used "means"
        instead of .gt. The comparison would just return False.
        """
        update = make_factor["f_irrelevant_3"].update_context_register(
            make_predicate["p2"],
            {str(make_entity["dan"]): make_entity["dan"]},
            operator.gt,
        )
        with pytest.raises(TypeError):
            any(register is not None for register in update)


class TestConsistent:
    def test_contradictory_facts_about_same_entity(self, watt_factor):
        left = watt_factor["f8_less"]
        right = watt_factor["f8_meters"]
        register = ContextRegister()
        register.insert_pair(left.generic_terms()[0], right.generic_terms()[0])
        assert not left.consistent_with(right, register)
        assert left.explain_consistent_with(right, register) is None

    def test_explanations_consistent_with(self, watt_factor):
        left = watt_factor["f8_less"]
        right = watt_factor["f8_meters"]
        register = ContextRegister()
        register.insert_pair(left.generic_terms()[0], right.generic_terms()[0])
        explanations = list(left.explanations_consistent_with(right, context=register))
        assert not explanations

    def test_factor_consistent_with_none(self, make_exhibit):
        assert make_exhibit["no_shooting_testimony"].consistent_with(
            make_exhibit["no_shooting_witness_unknown_testimony"]
        )


class TestAddition:
    @pytest.mark.parametrize(
        "left, right, expected",
        [
            ("f_shooting_craig_poe", "f_shooting_craig_brd", "f_shooting_craig_brd"),
            ("f_irrelevant_3", "f_irrelevant_3_new_context", "f_irrelevant_3"),
            (
                "f_irrelevant_3_new_context",
                "f_irrelevant_3",
                "f_irrelevant_3_new_context",
            ),
        ],
    )
    def test_addition(self, make_factor, left, right, expected):
        answer = make_factor[left] + make_factor[right]
        assert answer.means(make_factor[expected])

    def test_add_unrelated_factors(self, make_factor):
        assert make_factor["f_murder"] + make_factor["f_crime"] is None

    def test_cant_add_enactment_to_fact(self, watt_factor, e_search_clause):
        with pytest.raises(TypeError):
            print(watt_factor["f3"] + e_search_clause)



"""Classes describing judicial institutions and their effect on legal authority."""

from __future__ import annotations

from typing import List, Optional

from dataclasses import dataclass

from authorityspoke.opinions import Opinion


@dataclass
class Court:
    """
    An institution that issues :class:`.Opinion`\s and decides litigation.

    Has :class:`.Judge`\s and may be inferior to other :class:`Court`\s.

    :param id:
        an identifier

    :param opinions:
        a list of :class:`.Opinion`\s issued

    :param inferior_to:
        a :class:`Court` with the power to overrule
        :class:`.Rule`\s posited by this :class:`Court`
    """

    id: str
    opinions: Optional[List[Opinion]] = None
    inferior_to: Optional[List[Court]] = None



from authorityspoke.evidence import Evidence, Exhibit
from authorityspoke.facts import build_fact

from nettlesome.terms import TermSequence


class TestEvidence:
    def test_make_evidence_object(self, watt_factor):
        e = Evidence(Exhibit(form="testimony"), to_effect=watt_factor["f2"])
        assert not e.absent

    def test_default_len_based_on_unique_entity_slots(self, make_entity, make_factor):
        """same as e["no_shooting"]"""

        e = Evidence(
            Exhibit(
                form="testimony",
                statement=make_factor["f_no_shooting"],
                statement_attribution=make_entity["alice"],
            ),
            to_effect=make_factor["f_no_crime"],
        )
        assert not e.generic

    def test_no_extra_space_around_exhibit_in_string(self, make_opinion_with_holding):
        """
        Don't expect the holdings imported from the JSON to
        exactly match the holdings created for testing in conftest.
        """

        lotus = make_opinion_with_holding["lotus_majority"]
        holding = list(lotus.holdings)[4]
        assert " , " not in str(holding.inputs[0])

    def test_no_caps_in_short_string(self, make_opinion_with_holding):
        """
        Class names should not be capitalized in this format.
        """
        lotus = make_opinion_with_holding["lotus_majority"]
        evidence = list(lotus.holdings)[4].inputs[0]
        assert "Fact" not in evidence.short_string
        assert "fact" in evidence.short_string

    def test_get_entity_orders(self, make_evidence):
        context = make_evidence["no_shooting"].exhibit.statement.terms
        assert "Alice" in str(context[0])
        assert "Bob" in str(context[1])

    def test_get_entity_orders_no_statement(self, make_factor):
        e = Evidence(Exhibit(form="testimony"), to_effect=make_factor["f_no_crime"])
        assert len(e.to_effect.terms) == 1

    def test_evidence_str_with_context(self, make_evidence):
        assert (
            "in the form testimony"
            in make_evidence["reciprocal"].wrapped_string.lower()
        )

    def test_type_of_terms(self, make_evidence):
        assert isinstance(make_evidence["no_shooting"].terms, TermSequence)


class TestEvidenceSameMeaning:
    def test_equality_with_entity_order(self, make_evidence):
        e = make_evidence
        assert e["no_shooting"].means(e["no_shooting_entity_order"])

    def test_equality_with_no_statement(self, make_evidence):
        assert make_evidence["crime"].means(make_evidence["crime"])

    def test_unequal_due_to_entity_order(self, make_evidence):
        e = make_evidence
        assert not e["no_shooting"].means(e["no_shooting_different_witness"])

    def test_unequal_different_attributes(self, make_evidence):
        assert not (
            make_evidence["no_shooting_no_effect_entity_order"].means(
                make_evidence["no_shooting_different_witness"]
            )
        )

    def test_not_equal_no_effect(self, make_evidence):
        assert not make_evidence["shooting_no_effect"].means(make_evidence["shooting"])
        assert not make_evidence["shooting"].means(make_evidence["shooting_no_effect"])


class TestEvidenceImplication:
    def test_implication_missing_witness(self, make_evidence):
        e = make_evidence
        assert e["no_shooting"] >= e["no_shooting_witness_unknown"]

    def test_implication_missing_effect(self, make_evidence):
        e = make_evidence
        assert e["no_shooting"] >= e["no_shooting_no_effect_entity_order"]

    def test_no_implication_of_fact(
        self, make_predicate, make_evidence, watt_mentioned
    ):
        cool_fact = build_fact(
            make_predicate["p_no_shooting"], case_factors=watt_mentioned
        )
        assert not make_evidence["no_shooting"] > cool_fact
        assert not cool_fact > make_evidence["no_shooting"]


class TestEvidenceContradiction:
    def test_no_contradiction_of_fact(self, make_evidence, make_factor):
        assert not make_evidence["no_shooting"].contradicts(
            make_factor["f_no_shooting"]
        )

    def test_no_contradiction_from_supporting_contradictory_facts(self, make_evidence):
        assert not make_evidence["no_shooting"].contradicts(make_evidence["shooting"])

    def test_contradiction_of_absent_version_of_self(self, make_evidence):
        e = make_evidence
        assert e["no_shooting"].contradicts(e["no_shooting_absent"])

    def test_contradict_absent_version_of_implied_factor(self, make_evidence):
        e = make_evidence
        assert e["no_shooting_witness_unknown_absent"].contradicts(e["no_shooting"])
        assert e["no_shooting"].contradicts(e["no_shooting_witness_unknown_absent"])

    def test_no_contradiction_absent_same_witness(self, make_evidence):
        e = make_evidence
        assert not e["no_shooting_absent"].contradicts(e["no_shooting_witness_unknown"])
        assert not e["no_shooting_witness_unknown"].contradicts(e["no_shooting_absent"])
        assert not e["no_shooting_absent"].contradicts(
            e["no_shooting_different_witness"]
        )

    def test_no_contradiction_of_implied_factor(self, make_evidence):
        e = make_evidence
        assert not e["no_shooting"].contradicts(e["no_shooting_witness_unknown"])



"""AuthoritySpoke: Reading the law for the last time."""

from legislice import Enactment
from legislice.download import Client as LegisClient
from nettlesome.entities import Entity
from nettlesome.predicates import Predicate
from nettlesome.quantities import Comparison

from .facts import Fact
from .holdings import Holding
from .opinions import Opinion
from .rules import Rule
from .io.downloads import CAPClient
from .io.dump import to_dict, to_json

__version__ = "0.7.1"
